
#include <cdn-rawc/cdn-rawc-macros.h>

#ifndef CDN_MATH_DEFINE_PROTOS
#define CDN_MATH_DEFINE_PROTOS

#ifdef CDN_MATH_MATRIX_MULTIPLY_REQUIRED
#define CDN_MATH_MATRIX_MULTIPLY_V_REQUIRED
#endif

#include <math.h>
#include <stdlib.h>
#include <stdint.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define CDN_MATH_VALUE_TYPE_FUNC_REAL_ONE_MORE(Func,ValueType) CDN_MATH_VALUE_TYPE_FUNC_##ValueType(Func)
#define CDN_MATH_VALUE_TYPE_FUNC_REAL(Func,ValueType) CDN_MATH_VALUE_TYPE_FUNC_REAL_ONE_MORE(Func,ValueType)
#define CDN_MATH_VALUE_TYPE_FUNC(Func) CDN_MATH_VALUE_TYPE_FUNC_REAL(Func,ValueType)
#define CDN_MATH_VALUE_TYPE_FUNC_float(Func) Func##f
#define CDN_MATH_VALUE_TYPE_FUNC_double(Func) Func

#ifdef ENABLE_LAPACK
#ifdef PLATFORM_OSX
#define LP_int __CLPK_integer
#define LP_double __CLPK_doublereal
#define LP_float __CLPK_floatreal
#else
#define LP_int int32_t
#define LP_double double
#define LP_float float
#endif

#define LP_ValueTypeRealOneMore(ValueType) LP_##ValueType
#define LP_ValueTypeReal(ValueType) LP_ValueTypeRealOneMore(ValueType)
#define LP_ValueType LP_ValueTypeReal(ValueType)

#ifndef PLATFORM_OSX
extern void dgetrf_ (LP_int *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_int *,
                     LP_int *);

extern void dgetri_ (LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_int *);

extern void dgelsd_ (LP_int *,
                     LP_int *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_double *,
                     LP_double *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_int *,
                     LP_int *);

extern void dgesv_ (LP_int *,
                    LP_int *,
                    LP_double *,
                    LP_int *,
                    LP_int *,
                    LP_double *,
                    LP_int *,
                    LP_int *);

extern void dgeqrf_ (LP_int *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_double *,
                     LP_double *,
                     LP_int *,
                     LP_int *);

extern void dorgqr_ (LP_int *,
                     LP_int *,
                     LP_int *,
                     LP_double *,
                     LP_int *,
                     LP_double *,
                     LP_double *,
                     LP_int *,
                     LP_int *);



extern void sgetrf_ (LP_int *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_int *,
                     LP_int *);

extern void sgetri_ (LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_int *);

extern void sgelsd_ (LP_int *,
                     LP_int *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_float *,
                     LP_float *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_int *,
                     LP_int *);

extern void sgesv_ (LP_int *,
                    LP_int *,
                    LP_float *,
                    LP_int *,
                    LP_int *,
                    LP_float *,
                    LP_int *,
                    LP_int *);

extern void sgeqrf_ (LP_int *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_float *,
                     LP_float *,
                     LP_int *,
                     LP_int *);

extern void sorgqr_ (LP_int *,
                     LP_int *,
                     LP_int *,
                     LP_float *,
                     LP_int *,
                     LP_float *,
                     LP_float *,
                     LP_int *,
                     LP_int *);

#endif

#endif



#if (defined(CDN_MATH_SIN_V_REQUIRED) && !defined(CDN_MATH_SIN_V))
#define CDN_MATH_SIN_REQUIRED
#endif

#if (defined(CDN_MATH_COS_V_REQUIRED) && !defined(CDN_MATH_COS_V))
#define CDN_MATH_COS_REQUIRED
#endif

#if (defined(CDN_MATH_TAN_V_REQUIRED) && !defined(CDN_MATH_TAN_V))
#define CDN_MATH_TAN_REQUIRED
#endif

#if (defined(CDN_MATH_ASIN_V_REQUIRED) && !defined(CDN_MATH_ASIN_V))
#define CDN_MATH_ASIN_REQUIRED
#endif

#if (defined(CDN_MATH_ACOS_V_REQUIRED) && !defined(CDN_MATH_ACOS_V))
#define CDN_MATH_ACOS_REQUIRED
#endif

#if (defined(CDN_MATH_ATAN_V_REQUIRED) && !defined(CDN_MATH_ATAN_V))
#define CDN_MATH_ATAN_REQUIRED
#endif

#if (defined(CDN_MATH_SINH_V_REQUIRED) && !defined(CDN_MATH_SINH_V))
#define CDN_MATH_SINH_REQUIRED
#endif

#if (defined(CDN_MATH_COSH_V_REQUIRED) && !defined(CDN_MATH_COSH_V))
#define CDN_MATH_COSH_REQUIRED
#endif

#if (defined(CDN_MATH_TANH_V_REQUIRED) && !defined(CDN_MATH_TANH_V))
#define CDN_MATH_TANH_REQUIRED
#endif

#if (defined(CDN_MATH_SQRT_V_REQUIRED) && !defined(CDN_MATH_SQRT_V))
#define CDN_MATH_SQRT_REQUIRED
#endif

#if (defined(CDN_MATH_FLOOR_V_REQUIRED) && !defined(CDN_MATH_FLOOR_V))
#define CDN_MATH_FLOOR_REQUIRED
#endif

#if (defined(CDN_MATH_CEIL_V_REQUIRED) && !defined(CDN_MATH_CEIL_V))
#define CDN_MATH_CEIL_REQUIRED
#endif

#if (defined(CDN_MATH_ROUND_V_REQUIRED) && !defined(CDN_MATH_ROUND_V))
#define CDN_MATH_ROUND_REQUIRED
#endif

#if (defined(CDN_MATH_EXP_V_REQUIRED) && !defined(CDN_MATH_EXP_V))
#define CDN_MATH_EXP_REQUIRED
#endif

#if (defined(CDN_MATH_ERF_V_REQUIRED) && !defined(CDN_MATH_ERF_V))
#define CDN_MATH_ERF_REQUIRED
#endif

#if (defined(CDN_MATH_LOG10_V_REQUIRED) && !defined(CDN_MATH_LOG10_V))
#define CDN_MATH_LOG10_REQUIRED
#endif

#if (defined(CDN_MATH_EXP2_V_REQUIRED) && !defined(CDN_MATH_EXP2_V))
#define CDN_MATH_EXP2_REQUIRED
#endif

#if (defined(CDN_MATH_INVSQRT_V_REQUIRED) && !defined(CDN_MATH_INVSQRT_V))
#define CDN_MATH_INVSQRT_REQUIRED
#endif

#if (defined(CDN_MATH_ABS_V_REQUIRED) && !defined(CDN_MATH_ABS_V))
#define CDN_MATH_ABS_REQUIRED
#endif

#if (defined(CDN_MATH_LN_V_REQUIRED) && !defined(CDN_MATH_LN_V))
#define CDN_MATH_LN_REQUIRED
#endif

#if (defined(CDN_MATH_SIGN_V_REQUIRED) && !defined(CDN_MATH_SIGN_V))
#define CDN_MATH_SIGN_REQUIRED
#endif

#if (defined(CDN_MATH_ATAN2_V_M_M_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_M)) || \
   (defined(CDN_MATH_ATAN2_V_1_M_REQUIRED) && !defined(CDN_MATH_ATAN2_V_1_M)) || \
   (defined(CDN_MATH_ATAN2_V_M_1_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_1))
#define CDN_MATH_ATAN2_REQUIRED
#endif

#if (defined(CDN_MATH_POW_V_M_M_REQUIRED) && !defined(CDN_MATH_POW_V_M_M)) || \
   (defined(CDN_MATH_POW_V_1_M_REQUIRED) && !defined(CDN_MATH_POW_V_1_M)) || \
   (defined(CDN_MATH_POW_V_M_1_REQUIRED) && !defined(CDN_MATH_POW_V_M_1))
#define CDN_MATH_POW_REQUIRED
#endif

#if (defined(CDN_MATH_HYPOT_V_M_M_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_M)) || \
   (defined(CDN_MATH_HYPOT_V_1_M_REQUIRED) && !defined(CDN_MATH_HYPOT_V_1_M)) || \
   (defined(CDN_MATH_HYPOT_V_M_1_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_1))
#define CDN_MATH_HYPOT_REQUIRED
#endif

#if (defined(CDN_MATH_MAX_V_M_M_REQUIRED) && !defined(CDN_MATH_MAX_V_M_M)) || \
   (defined(CDN_MATH_MAX_V_1_M_REQUIRED) && !defined(CDN_MATH_MAX_V_1_M)) || \
   (defined(CDN_MATH_MAX_V_M_1_REQUIRED) && !defined(CDN_MATH_MAX_V_M_1))
#define CDN_MATH_MAX_REQUIRED
#endif

#if (defined(CDN_MATH_MIN_V_M_M_REQUIRED) && !defined(CDN_MATH_MIN_V_M_M)) || \
   (defined(CDN_MATH_MIN_V_1_M_REQUIRED) && !defined(CDN_MATH_MIN_V_1_M)) || \
   (defined(CDN_MATH_MIN_V_M_1_REQUIRED) && !defined(CDN_MATH_MIN_V_M_1))
#define CDN_MATH_MIN_REQUIRED
#endif

#if (defined(CDN_MATH_MODULO_V_M_M_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_M)) || \
   (defined(CDN_MATH_MODULO_V_1_M_REQUIRED) && !defined(CDN_MATH_MODULO_V_1_M)) || \
   (defined(CDN_MATH_MODULO_V_M_1_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_1))
#define CDN_MATH_MODULO_REQUIRED
#endif

#if (defined(CDN_MATH_CSIGN_V_M_M_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_M)) || \
   (defined(CDN_MATH_CSIGN_V_1_M_REQUIRED) && !defined(CDN_MATH_CSIGN_V_1_M)) || \
   (defined(CDN_MATH_CSIGN_V_M_1_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_1))
#define CDN_MATH_CSIGN_REQUIRED
#endif

#if (defined(CDN_MATH_SUM_V_M_M_REQUIRED) && !defined(CDN_MATH_SUM_V_M_M)) || \
   (defined(CDN_MATH_SUM_V_1_M_REQUIRED) && !defined(CDN_MATH_SUM_V_1_M)) || \
   (defined(CDN_MATH_SUM_V_M_1_REQUIRED) && !defined(CDN_MATH_SUM_V_M_1))
#define CDN_MATH_SUM_REQUIRED
#endif

#if (defined(CDN_MATH_PRODUCT_V_M_M_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_M)) || \
   (defined(CDN_MATH_PRODUCT_V_1_M_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_1_M)) || \
   (defined(CDN_MATH_PRODUCT_V_M_1_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_1))
#define CDN_MATH_PRODUCT_REQUIRED
#endif

#if (defined(CDN_MATH_SQSUM_V_M_M_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_M)) || \
   (defined(CDN_MATH_SQSUM_V_1_M_REQUIRED) && !defined(CDN_MATH_SQSUM_V_1_M)) || \
   (defined(CDN_MATH_SQSUM_V_M_1_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_1))
#define CDN_MATH_SQSUM_REQUIRED
#endif

#if (defined(CDN_MATH_LERP_V_M_M_M_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_M)) || \
   (defined(CDN_MATH_LERP_V_M_M_1_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_1)) || \
   (defined(CDN_MATH_LERP_V_M_1_M_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_M)) || \
   (defined(CDN_MATH_LERP_V_M_1_1_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_1)) || \
   (defined(CDN_MATH_LERP_V_1_M_M_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_M)) || \
   (defined(CDN_MATH_LERP_V_1_M_1_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_1)) || \
   (defined(CDN_MATH_LERP_V_1_1_M_REQUIRED) && !defined(CDN_MATH_LERP_V_1_1_M))
#define CDN_MATH_LERP_REQUIRED
#endif

#if (defined(CDN_MATH_CLIP_V_M_M_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_M)) || \
   (defined(CDN_MATH_CLIP_V_M_M_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_1)) || \
   (defined(CDN_MATH_CLIP_V_M_1_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_M)) || \
   (defined(CDN_MATH_CLIP_V_M_1_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_1)) || \
   (defined(CDN_MATH_CLIP_V_1_M_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_M)) || \
   (defined(CDN_MATH_CLIP_V_1_M_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_1)) || \
   (defined(CDN_MATH_CLIP_V_1_1_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_1_M))
#define CDN_MATH_CLIP_REQUIRED
#endif

#if (defined(CDN_MATH_CYCLE_V_M_M_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_M)) || \
   (defined(CDN_MATH_CYCLE_V_M_M_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_1)) || \
   (defined(CDN_MATH_CYCLE_V_M_1_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_M)) || \
   (defined(CDN_MATH_CYCLE_V_M_1_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_1)) || \
   (defined(CDN_MATH_CYCLE_V_1_M_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_M)) || \
   (defined(CDN_MATH_CYCLE_V_1_M_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_1)) || \
   (defined(CDN_MATH_CYCLE_V_1_1_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_1_M))
#define CDN_MATH_CYCLE_REQUIRED
#endif

#if (defined(CDN_MATH_MAX_V_REQUIRED) && !defined(CDN_MATH_MAX_V))
#define CDN_MATH_MAX_REQUIRED
#endif

#if (defined(CDN_MATH_MIN_V_REQUIRED) && !defined(CDN_MATH_MIN_V))
#define CDN_MATH_MIN_REQUIRED
#endif

#if (defined(CDN_MATH_HYPOT_V_REQUIRED) && !defined(CDN_MATH_HYPOT_V))
#define CDN_MATH_SQRT_REQUIRED
#endif

#if defined(CDN_MATH_SIN_REQUIRED) && !defined(CDN_MATH_SIN)
#define CDN_MATH_SIN_USE_BUILTIN
#define CDN_MATH_SIN cdn_math_sin_builtin

static inline ValueType cdn_math_sin_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_sin_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(sin)(x0);
}
#endif /* CDN_MATH_SIN */

#if defined(CDN_MATH_COS_REQUIRED) && !defined(CDN_MATH_COS)
#define CDN_MATH_COS_USE_BUILTIN
#define CDN_MATH_COS cdn_math_cos_builtin

static inline ValueType cdn_math_cos_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_cos_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(cos)(x0);
}
#endif /* CDN_MATH_COS */

#if defined(CDN_MATH_TAN_REQUIRED) && !defined(CDN_MATH_TAN)
#define CDN_MATH_TAN_USE_BUILTIN
#define CDN_MATH_TAN cdn_math_tan_builtin

static inline ValueType cdn_math_tan_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_tan_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(tan)(x0);
}
#endif /* CDN_MATH_TAN */

#if defined(CDN_MATH_ASIN_REQUIRED) && !defined(CDN_MATH_ASIN)
#define CDN_MATH_ASIN_USE_BUILTIN
#define CDN_MATH_ASIN cdn_math_asin_builtin

static inline ValueType cdn_math_asin_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_asin_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(asin)(x0);
}
#endif /* CDN_MATH_ASIN */

#if defined(CDN_MATH_ACOS_REQUIRED) && !defined(CDN_MATH_ACOS)
#define CDN_MATH_ACOS_USE_BUILTIN
#define CDN_MATH_ACOS cdn_math_acos_builtin

static inline ValueType cdn_math_acos_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_acos_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(acos)(x0);
}
#endif /* CDN_MATH_ACOS */

#if defined(CDN_MATH_ATAN_REQUIRED) && !defined(CDN_MATH_ATAN)
#define CDN_MATH_ATAN_USE_BUILTIN
#define CDN_MATH_ATAN cdn_math_atan_builtin

static inline ValueType cdn_math_atan_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_atan_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(atan)(x0);
}
#endif /* CDN_MATH_ATAN */

#if defined(CDN_MATH_SINH_REQUIRED) && !defined(CDN_MATH_SINH)
#define CDN_MATH_SINH_USE_BUILTIN
#define CDN_MATH_SINH cdn_math_sinh_builtin

static inline ValueType cdn_math_sinh_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_sinh_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(sinh)(x0);
}
#endif /* CDN_MATH_SINH */

#if defined(CDN_MATH_COSH_REQUIRED) && !defined(CDN_MATH_COSH)
#define CDN_MATH_COSH_USE_BUILTIN
#define CDN_MATH_COSH cdn_math_cosh_builtin

static inline ValueType cdn_math_cosh_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_cosh_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(cosh)(x0);
}
#endif /* CDN_MATH_COSH */

#if defined(CDN_MATH_TANH_REQUIRED) && !defined(CDN_MATH_TANH)
#define CDN_MATH_TANH_USE_BUILTIN
#define CDN_MATH_TANH cdn_math_tanh_builtin

static inline ValueType cdn_math_tanh_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_tanh_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(tanh)(x0);
}
#endif /* CDN_MATH_TANH */

#if defined(CDN_MATH_SQRT_REQUIRED) && !defined(CDN_MATH_SQRT)
#define CDN_MATH_SQRT_USE_BUILTIN
#define CDN_MATH_SQRT cdn_math_sqrt_builtin

static inline ValueType cdn_math_sqrt_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_sqrt_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(sqrt)(x0);
}
#endif /* CDN_MATH_SQRT */

#if defined(CDN_MATH_FLOOR_REQUIRED) && !defined(CDN_MATH_FLOOR)
#define CDN_MATH_FLOOR_USE_BUILTIN
#define CDN_MATH_FLOOR cdn_math_floor_builtin

static inline ValueType cdn_math_floor_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_floor_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(floor)(x0);
}
#endif /* CDN_MATH_FLOOR */

#if defined(CDN_MATH_CEIL_REQUIRED) && !defined(CDN_MATH_CEIL)
#define CDN_MATH_CEIL_USE_BUILTIN
#define CDN_MATH_CEIL cdn_math_ceil_builtin

static inline ValueType cdn_math_ceil_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_ceil_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(ceil)(x0);
}
#endif /* CDN_MATH_CEIL */

#if defined(CDN_MATH_ROUND_REQUIRED) && !defined(CDN_MATH_ROUND)
#define CDN_MATH_ROUND_USE_BUILTIN
#define CDN_MATH_ROUND cdn_math_round_builtin

static inline ValueType cdn_math_round_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_round_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(round)(x0);
}
#endif /* CDN_MATH_ROUND */

#if defined(CDN_MATH_EXP_REQUIRED) && !defined(CDN_MATH_EXP)
#define CDN_MATH_EXP_USE_BUILTIN
#define CDN_MATH_EXP cdn_math_exp_builtin

static inline ValueType cdn_math_exp_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_exp_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(exp)(x0);
}
#endif /* CDN_MATH_EXP */

#if defined(CDN_MATH_ERF_REQUIRED) && !defined(CDN_MATH_ERF)
#define CDN_MATH_ERF_USE_BUILTIN
#define CDN_MATH_ERF cdn_math_erf_builtin

static inline ValueType cdn_math_erf_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_erf_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(erf)(x0);
}
#endif /* CDN_MATH_ERF */

#if defined(CDN_MATH_LOG10_REQUIRED) && !defined(CDN_MATH_LOG10)
#define CDN_MATH_LOG10_USE_BUILTIN
#define CDN_MATH_LOG10 cdn_math_log10_builtin

static inline ValueType cdn_math_log10_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_log10_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(log10)(x0);
}
#endif /* CDN_MATH_LOG10 */

#if defined(CDN_MATH_EXP2_REQUIRED) && !defined(CDN_MATH_EXP2)
#define CDN_MATH_EXP2_USE_BUILTIN
#define CDN_MATH_EXP2 cdn_math_exp2_builtin

static inline ValueType cdn_math_exp2_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_exp2_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(exp2)(x0);
}
#endif /* CDN_MATH_EXP2 */

#if defined(CDN_MATH_ATAN2_REQUIRED) && !defined(CDN_MATH_ATAN2)
#define CDN_MATH_ATAN2_USE_BUILTIN
#define CDN_MATH_ATAN2 cdn_math_atan2_builtin

static inline ValueType cdn_math_atan2_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_atan2_builtin (ValueType x0, ValueType x1)
{
	return CDN_MATH_VALUE_TYPE_FUNC(atan2)(x0, x1);
}
#endif /* CDN_MATH_ATAN2 */

#if defined(CDN_MATH_POW_REQUIRED) && !defined(CDN_MATH_POW)
#define CDN_MATH_POW_USE_BUILTIN
#define CDN_MATH_POW cdn_math_pow_builtin

static inline ValueType cdn_math_pow_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_pow_builtin (ValueType x0, ValueType x1)
{
	return CDN_MATH_VALUE_TYPE_FUNC(pow)(x0, x1);
}
#endif /* CDN_MATH_POW */

#if defined(CDN_MATH_HYPOT_REQUIRED) && !defined(CDN_MATH_HYPOT)
#define CDN_MATH_HYPOT_USE_BUILTIN
#define CDN_MATH_HYPOT cdn_math_hypot_builtin

static inline ValueType cdn_math_hypot_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_hypot_builtin (ValueType x0, ValueType x1)
{
	return CDN_MATH_VALUE_TYPE_FUNC(hypot)(x0, x1);
}
#endif /* CDN_MATH_HYPOT */

#if defined(CDN_MATH_INVSQRT_REQUIRED) && !defined(CDN_MATH_INVSQRT)
#define CDN_MATH_INVSQRT_USE_BUILTIN
#define CDN_MATH_INVSQRT cdn_math_invsqrt_builtin

static inline ValueType cdn_math_invsqrt_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_invsqrt_builtin (ValueType x0)
{
	return 1.0 / CDN_MATH_SQRT (x0);
}
#endif /* CDN_MATH_INVSQRT */

#if defined(CDN_MATH_ABS_REQUIRED) && !defined(CDN_MATH_ABS)
#define CDN_MATH_ABS_USE_BUILTIN
#define CDN_MATH_ABS cdn_math_abs_builtin

static inline ValueType cdn_math_abs_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_abs_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(fabs) (x0);
}
#endif /* CDN_MATH_ABS */

#if defined(CDN_MATH_LN_REQUIRED) && !defined(CDN_MATH_LN)
#define CDN_MATH_LN_USE_BUILTIN
#define CDN_MATH_LN cdn_math_ln_builtin

static inline ValueType cdn_math_ln_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_ln_builtin (ValueType x0)
{
	return CDN_MATH_VALUE_TYPE_FUNC(log) (x0);
}
#endif /* CDN_MATH_LN */

#if defined(CDN_MATH_SIGN_REQUIRED) && !defined(CDN_MATH_SIGN)
#define CDN_MATH_SIGN_USE_BUILTIN
#define CDN_MATH_SIGN cdn_math_sign_builtin

static inline ValueType cdn_math_sign_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_sign_builtin (ValueType x0)
{
	return signbit (x0) ? -1 : 1;
}
#endif /* CDN_MATH_SIGN */

#if defined(CDN_MATH_CSIGN_REQUIRED) && !defined(CDN_MATH_CSIGN)
#define CDN_MATH_CSIGN_USE_BUILTIN
#define CDN_MATH_CSIGN cdn_math_csign_builtin

static inline ValueType cdn_math_csign_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_csign_builtin (ValueType x0, ValueType x1)
{
	return copysign (x0, x1);
}
#endif /* CDN_MATH_CSIGN */

#if defined(CDN_MATH_LERP_REQUIRED) && !defined(CDN_MATH_LERP)
#define CDN_MATH_LERP_USE_BUILTIN
#define CDN_MATH_LERP cdn_math_lerp_builtin

static inline ValueType cdn_math_lerp_builtin (ValueType x0, ValueType x1, ValueType x2) GNUC_INLINE;

static ValueType cdn_math_lerp_builtin (ValueType x0, ValueType x1, ValueType x2)
{
	return x1 + (x2 - x1) * x0;
}
#endif /* CDN_MATH_LERP */

#if defined(CDN_MATH_CLIP_REQUIRED) && !defined(CDN_MATH_CLIP)
#define CDN_MATH_CLIP_USE_BUILTIN
#define CDN_MATH_CLIP cdn_math_clip_builtin

static inline ValueType cdn_math_clip_builtin (ValueType x0, ValueType x1, ValueType x2) GNUC_INLINE;

static ValueType cdn_math_clip_builtin (ValueType x0, ValueType x1, ValueType x2)
{
	if (x0 < x1)
	{
		return x1;
	}
	else if (x0 > x2)
	{
		return x2;
	}
	else
	{
		return x0;
	}
}
#endif /* CDN_MATH_CLIP */

#if defined(CDN_MATH_CYCLE_REQUIRED) && !defined(CDN_MATH_CYCLE)
#define CDN_MATH_CYCLE_USE_BUILTIN
#define CDN_MATH_CYCLE cdn_math_cycle_builtin

static inline ValueType cdn_math_cycle_builtin (ValueType x0, ValueType x1, ValueType x2) GNUC_INLINE;

static ValueType cdn_math_cycle_builtin (ValueType x0, ValueType x1, ValueType x2)
{
	if (x0 < x1)
	{
		return x2 - CDN_MATH_VALUE_TYPE_FUNC(fmod) (x1 - x0, x2 - x1);
	}
	else if (x0 > x2)
	{
		return x1 + CDN_MATH_VALUE_TYPE_FUNC(fmod) (x0 - x1, x2 - x1);
	}
	else
	{
		return x0;
	}
}
#endif /* CDN_MATH_CYCLE */

#if defined(CDN_MATH_MIN_REQUIRED) && !defined(CDN_MATH_MIN)
#define CDN_MATH_MIN_USE_BUILTIN
#define CDN_MATH_MIN cdn_math_min_builtin

static inline ValueType cdn_math_min_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_min_builtin (ValueType x0, ValueType x1)
{
	return x0 < x1 ? x0 : x1;
}
#endif /* CDN_MATH_MIN */

#if defined(CDN_MATH_MAX_REQUIRED) && !defined(CDN_MATH_MAX)
#define CDN_MATH_MAX_USE_BUILTIN
#define CDN_MATH_MAX cdn_math_max_builtin

static inline ValueType cdn_math_max_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_max_builtin (ValueType x0, ValueType x1)
{
	return x0 > x1 ? x0 : x1;
}
#endif /* CDN_MATH_MAX */

#if defined(CDN_MATH_HYPOT_REQUIRED) && !defined(CDN_MATH_HYPOT)
#define CDN_MATH_HYPOT_USE_BUILTIN
#define CDN_MATH_HYPOT cdn_math_hypot_builtin

static inline ValueType cdn_math_hypot_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_hypot_builtin (ValueType x0, ValueType x1)
{
	return hypot(x0, x1);
}
#endif /* CDN_MATH_HYPOT */

#if defined(CDN_MATH_PRODUCT_REQUIRED) && !defined(CDN_MATH_PRODUCT)
#define CDN_MATH_PRODUCT_USE_BUILTIN
#define CDN_MATH_PRODUCT cdn_math_product_builtin

static inline ValueType cdn_math_product_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_product_builtin (ValueType x0, ValueType x1)
{
	return x0 * x1;
}
#endif /* CDN_MATH_PRODUCT */

#if defined(CDN_MATH_SUM_REQUIRED) && !defined(CDN_MATH_SUM)
#define CDN_MATH_SUM_USE_BUILTIN
#define CDN_MATH_SUM cdn_math_sum_builtin

static inline ValueType cdn_math_sum_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_sum_builtin (ValueType x0, ValueType x1)
{
	return x0 + x1;
}
#endif /* CDN_MATH_SUM */

#if defined(CDN_MATH_SQSUM_REQUIRED) && !defined(CDN_MATH_SQSUM)
#define CDN_MATH_SQSUM_USE_BUILTIN
#define CDN_MATH_SQSUM cdn_math_sqsum_builtin

static inline ValueType cdn_math_sqsum_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_sqsum_builtin (ValueType x0, ValueType x1)
{
	return x0 * x0 + x1 * x1;
}
#endif /* CDN_MATH_SQSUM */

#if defined(CDN_MATH_SQSUM_1_REQUIRED) && !defined(CDN_MATH_SQSUM_1)
#define CDN_MATH_SQSUM_1_USE_BUILTIN
#define CDN_MATH_SQSUM_1 cdn_math_sqsum_1_builtin

static inline ValueType cdn_math_sqsum_1_builtin (ValueType x0) GNUC_INLINE;

static ValueType cdn_math_sqsum_1_builtin (ValueType x0)
{
	return x0 * x0;
}
#endif /* CDN_MATH_SQSUM_1 */

#if defined(CDN_MATH_RAND_REQUIRED) && !defined(CDN_MATH_RAND)
#define CDN_MATH_RAND_USE_BUILTIN
#define CDN_MATH_RAND cdn_math_rand_builtin

static inline ValueType cdn_math_rand_builtin (void) GNUC_INLINE;

static ValueType cdn_math_rand_builtin (void)
{
	return (random () / (ValueType)RAND_MAX);
}
#endif /* CDN_MATH_RAND */

#if defined(CDN_MATH_MODULO_REQUIRED) && !defined(CDN_MATH_MODULO)
#define CDN_MATH_MODULO_USE_BUILTIN
#define CDN_MATH_MODULO cdn_math_modulo_builtin

static inline ValueType cdn_math_modulo_builtin (ValueType x0, ValueType x1) GNUC_INLINE;

static ValueType cdn_math_modulo_builtin (ValueType x0, ValueType x1)
{
	ValueType ans = CDN_MATH_VALUE_TYPE_FUNC(fmod) (x0, x1);

	return ans < 0 ? ans + x1 : ans;
}
#endif /* CDN_MATH_MODULO */

#if defined(CDN_MATH_SIN_V_REQUIRED) && !defined(CDN_MATH_SIN_V)
#define CDN_MATH_SIN_V_USE_BUILTIN
#define CDN_MATH_SIN_V cdn_math_sin_v_builtin

static ValueType *cdn_math_sin_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_sin_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SIN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SIN_V */

#if defined(CDN_MATH_SIN_V_IP_REQUIRED) && !defined(CDN_MATH_SIN_V_IP)
#define CDN_MATH_SIN_V_IP_USE_BUILTIN
#define CDN_MATH_SIN_V_IP cdn_math_sin_v_ip_builtin

static ValueType *cdn_math_sin_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_sin_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SIN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SIN_V_IP */

#if defined(CDN_MATH_COS_V_REQUIRED) && !defined(CDN_MATH_COS_V)
#define CDN_MATH_COS_V_USE_BUILTIN
#define CDN_MATH_COS_V cdn_math_cos_v_builtin

static ValueType *cdn_math_cos_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_cos_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_COS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_COS_V */

#if defined(CDN_MATH_COS_V_IP_REQUIRED) && !defined(CDN_MATH_COS_V_IP)
#define CDN_MATH_COS_V_IP_USE_BUILTIN
#define CDN_MATH_COS_V_IP cdn_math_cos_v_ip_builtin

static ValueType *cdn_math_cos_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_cos_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_COS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_COS_V_IP */

#if defined(CDN_MATH_TAN_V_REQUIRED) && !defined(CDN_MATH_TAN_V)
#define CDN_MATH_TAN_V_USE_BUILTIN
#define CDN_MATH_TAN_V cdn_math_tan_v_builtin

static ValueType *cdn_math_tan_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_tan_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_TAN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_TAN_V */

#if defined(CDN_MATH_TAN_V_IP_REQUIRED) && !defined(CDN_MATH_TAN_V_IP)
#define CDN_MATH_TAN_V_IP_USE_BUILTIN
#define CDN_MATH_TAN_V_IP cdn_math_tan_v_ip_builtin

static ValueType *cdn_math_tan_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_tan_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_TAN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_TAN_V_IP */

#if defined(CDN_MATH_ASIN_V_REQUIRED) && !defined(CDN_MATH_ASIN_V)
#define CDN_MATH_ASIN_V_USE_BUILTIN
#define CDN_MATH_ASIN_V cdn_math_asin_v_builtin

static ValueType *cdn_math_asin_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_asin_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ASIN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ASIN_V */

#if defined(CDN_MATH_ASIN_V_IP_REQUIRED) && !defined(CDN_MATH_ASIN_V_IP)
#define CDN_MATH_ASIN_V_IP_USE_BUILTIN
#define CDN_MATH_ASIN_V_IP cdn_math_asin_v_ip_builtin

static ValueType *cdn_math_asin_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_asin_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ASIN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ASIN_V_IP */

#if defined(CDN_MATH_ACOS_V_REQUIRED) && !defined(CDN_MATH_ACOS_V)
#define CDN_MATH_ACOS_V_USE_BUILTIN
#define CDN_MATH_ACOS_V cdn_math_acos_v_builtin

static ValueType *cdn_math_acos_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_acos_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ACOS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ACOS_V */

#if defined(CDN_MATH_ACOS_V_IP_REQUIRED) && !defined(CDN_MATH_ACOS_V_IP)
#define CDN_MATH_ACOS_V_IP_USE_BUILTIN
#define CDN_MATH_ACOS_V_IP cdn_math_acos_v_ip_builtin

static ValueType *cdn_math_acos_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_acos_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ACOS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ACOS_V_IP */

#if defined(CDN_MATH_ATAN_V_REQUIRED) && !defined(CDN_MATH_ATAN_V)
#define CDN_MATH_ATAN_V_USE_BUILTIN
#define CDN_MATH_ATAN_V cdn_math_atan_v_builtin

static ValueType *cdn_math_atan_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_atan_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ATAN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN_V */

#if defined(CDN_MATH_ATAN_V_IP_REQUIRED) && !defined(CDN_MATH_ATAN_V_IP)
#define CDN_MATH_ATAN_V_IP_USE_BUILTIN
#define CDN_MATH_ATAN_V_IP cdn_math_atan_v_ip_builtin

static ValueType *cdn_math_atan_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_atan_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ATAN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN_V_IP */

#if defined(CDN_MATH_SINH_V_REQUIRED) && !defined(CDN_MATH_SINH_V)
#define CDN_MATH_SINH_V_USE_BUILTIN
#define CDN_MATH_SINH_V cdn_math_sinh_v_builtin

static ValueType *cdn_math_sinh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_sinh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SINH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SINH_V */

#if defined(CDN_MATH_SINH_V_IP_REQUIRED) && !defined(CDN_MATH_SINH_V_IP)
#define CDN_MATH_SINH_V_IP_USE_BUILTIN
#define CDN_MATH_SINH_V_IP cdn_math_sinh_v_ip_builtin

static ValueType *cdn_math_sinh_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_sinh_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SINH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SINH_V_IP */

#if defined(CDN_MATH_COSH_V_REQUIRED) && !defined(CDN_MATH_COSH_V)
#define CDN_MATH_COSH_V_USE_BUILTIN
#define CDN_MATH_COSH_V cdn_math_cosh_v_builtin

static ValueType *cdn_math_cosh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_cosh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_COSH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_COSH_V */

#if defined(CDN_MATH_COSH_V_IP_REQUIRED) && !defined(CDN_MATH_COSH_V_IP)
#define CDN_MATH_COSH_V_IP_USE_BUILTIN
#define CDN_MATH_COSH_V_IP cdn_math_cosh_v_ip_builtin

static ValueType *cdn_math_cosh_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_cosh_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_COSH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_COSH_V_IP */

#if defined(CDN_MATH_TANH_V_REQUIRED) && !defined(CDN_MATH_TANH_V)
#define CDN_MATH_TANH_V_USE_BUILTIN
#define CDN_MATH_TANH_V cdn_math_tanh_v_builtin

static ValueType *cdn_math_tanh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_tanh_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_TANH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_TANH_V */

#if defined(CDN_MATH_TANH_V_IP_REQUIRED) && !defined(CDN_MATH_TANH_V_IP)
#define CDN_MATH_TANH_V_IP_USE_BUILTIN
#define CDN_MATH_TANH_V_IP cdn_math_tanh_v_ip_builtin

static ValueType *cdn_math_tanh_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_tanh_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_TANH (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_TANH_V_IP */

#if defined(CDN_MATH_SQRT_V_REQUIRED) && !defined(CDN_MATH_SQRT_V)
#define CDN_MATH_SQRT_V_USE_BUILTIN
#define CDN_MATH_SQRT_V cdn_math_sqrt_v_builtin

static ValueType *cdn_math_sqrt_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_sqrt_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SQRT (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQRT_V */

#if defined(CDN_MATH_SQRT_V_IP_REQUIRED) && !defined(CDN_MATH_SQRT_V_IP)
#define CDN_MATH_SQRT_V_IP_USE_BUILTIN
#define CDN_MATH_SQRT_V_IP cdn_math_sqrt_v_ip_builtin

static ValueType *cdn_math_sqrt_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_sqrt_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SQRT (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQRT_V_IP */

#if defined(CDN_MATH_FLOOR_V_REQUIRED) && !defined(CDN_MATH_FLOOR_V)
#define CDN_MATH_FLOOR_V_USE_BUILTIN
#define CDN_MATH_FLOOR_V cdn_math_floor_v_builtin

static ValueType *cdn_math_floor_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_floor_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_FLOOR (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_FLOOR_V */

#if defined(CDN_MATH_FLOOR_V_IP_REQUIRED) && !defined(CDN_MATH_FLOOR_V_IP)
#define CDN_MATH_FLOOR_V_IP_USE_BUILTIN
#define CDN_MATH_FLOOR_V_IP cdn_math_floor_v_ip_builtin

static ValueType *cdn_math_floor_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_floor_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_FLOOR (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_FLOOR_V_IP */

#if defined(CDN_MATH_CEIL_V_REQUIRED) && !defined(CDN_MATH_CEIL_V)
#define CDN_MATH_CEIL_V_USE_BUILTIN
#define CDN_MATH_CEIL_V cdn_math_ceil_v_builtin

static ValueType *cdn_math_ceil_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_ceil_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CEIL (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CEIL_V */

#if defined(CDN_MATH_CEIL_V_IP_REQUIRED) && !defined(CDN_MATH_CEIL_V_IP)
#define CDN_MATH_CEIL_V_IP_USE_BUILTIN
#define CDN_MATH_CEIL_V_IP cdn_math_ceil_v_ip_builtin

static ValueType *cdn_math_ceil_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_ceil_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CEIL (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CEIL_V_IP */

#if defined(CDN_MATH_ROUND_V_REQUIRED) && !defined(CDN_MATH_ROUND_V)
#define CDN_MATH_ROUND_V_USE_BUILTIN
#define CDN_MATH_ROUND_V cdn_math_round_v_builtin

static ValueType *cdn_math_round_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_round_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ROUND (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ROUND_V */

#if defined(CDN_MATH_ROUND_V_IP_REQUIRED) && !defined(CDN_MATH_ROUND_V_IP)
#define CDN_MATH_ROUND_V_IP_USE_BUILTIN
#define CDN_MATH_ROUND_V_IP cdn_math_round_v_ip_builtin

static ValueType *cdn_math_round_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_round_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ROUND (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ROUND_V_IP */

#if defined(CDN_MATH_EXP_V_REQUIRED) && !defined(CDN_MATH_EXP_V)
#define CDN_MATH_EXP_V_USE_BUILTIN
#define CDN_MATH_EXP_V cdn_math_exp_v_builtin

static ValueType *cdn_math_exp_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_exp_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_EXP (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_EXP_V */

#if defined(CDN_MATH_EXP_V_IP_REQUIRED) && !defined(CDN_MATH_EXP_V_IP)
#define CDN_MATH_EXP_V_IP_USE_BUILTIN
#define CDN_MATH_EXP_V_IP cdn_math_exp_v_ip_builtin

static ValueType *cdn_math_exp_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_exp_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_EXP (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_EXP_V_IP */

#if defined(CDN_MATH_ERF_V_REQUIRED) && !defined(CDN_MATH_ERF_V)
#define CDN_MATH_ERF_V_USE_BUILTIN
#define CDN_MATH_ERF_V cdn_math_erf_v_builtin

static ValueType *cdn_math_erf_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_erf_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ERF (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ERF_V */

#if defined(CDN_MATH_ERF_V_IP_REQUIRED) && !defined(CDN_MATH_ERF_V_IP)
#define CDN_MATH_ERF_V_IP_USE_BUILTIN
#define CDN_MATH_ERF_V_IP cdn_math_erf_v_ip_builtin

static ValueType *cdn_math_erf_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_erf_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ERF (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ERF_V_IP */

#if defined(CDN_MATH_LOG10_V_REQUIRED) && !defined(CDN_MATH_LOG10_V)
#define CDN_MATH_LOG10_V_USE_BUILTIN
#define CDN_MATH_LOG10_V cdn_math_log10_v_builtin

static ValueType *cdn_math_log10_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_log10_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LOG10 (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LOG10_V */

#if defined(CDN_MATH_LOG10_V_IP_REQUIRED) && !defined(CDN_MATH_LOG10_V_IP)
#define CDN_MATH_LOG10_V_IP_USE_BUILTIN
#define CDN_MATH_LOG10_V_IP cdn_math_log10_v_ip_builtin

static ValueType *cdn_math_log10_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_log10_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LOG10 (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LOG10_V_IP */

#if defined(CDN_MATH_EXP2_V_REQUIRED) && !defined(CDN_MATH_EXP2_V)
#define CDN_MATH_EXP2_V_USE_BUILTIN
#define CDN_MATH_EXP2_V cdn_math_exp2_v_builtin

static ValueType *cdn_math_exp2_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_exp2_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_EXP2 (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_EXP2_V */

#if defined(CDN_MATH_EXP2_V_IP_REQUIRED) && !defined(CDN_MATH_EXP2_V_IP)
#define CDN_MATH_EXP2_V_IP_USE_BUILTIN
#define CDN_MATH_EXP2_V_IP cdn_math_exp2_v_ip_builtin

static ValueType *cdn_math_exp2_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_exp2_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_EXP2 (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_EXP2_V_IP */

#if defined(CDN_MATH_INVSQRT_V_REQUIRED) && !defined(CDN_MATH_INVSQRT_V)
#define CDN_MATH_INVSQRT_V_USE_BUILTIN
#define CDN_MATH_INVSQRT_V cdn_math_invsqrt_v_builtin

static ValueType *cdn_math_invsqrt_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_invsqrt_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_INVSQRT (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_INVSQRT_V */

#if defined(CDN_MATH_INVSQRT_V_IP_REQUIRED) && !defined(CDN_MATH_INVSQRT_V_IP)
#define CDN_MATH_INVSQRT_V_IP_USE_BUILTIN
#define CDN_MATH_INVSQRT_V_IP cdn_math_invsqrt_v_ip_builtin

static ValueType *cdn_math_invsqrt_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_invsqrt_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_INVSQRT (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_INVSQRT_V_IP */

#if defined(CDN_MATH_ABS_V_REQUIRED) && !defined(CDN_MATH_ABS_V)
#define CDN_MATH_ABS_V_USE_BUILTIN
#define CDN_MATH_ABS_V cdn_math_abs_v_builtin

static ValueType *cdn_math_abs_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_abs_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ABS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ABS_V */

#if defined(CDN_MATH_ABS_V_IP_REQUIRED) && !defined(CDN_MATH_ABS_V_IP)
#define CDN_MATH_ABS_V_IP_USE_BUILTIN
#define CDN_MATH_ABS_V_IP cdn_math_abs_v_ip_builtin

static ValueType *cdn_math_abs_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_abs_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ABS (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ABS_V_IP */

#if defined(CDN_MATH_LN_V_REQUIRED) && !defined(CDN_MATH_LN_V)
#define CDN_MATH_LN_V_USE_BUILTIN
#define CDN_MATH_LN_V cdn_math_ln_v_builtin

static ValueType *cdn_math_ln_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_ln_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LN_V */

#if defined(CDN_MATH_LN_V_IP_REQUIRED) && !defined(CDN_MATH_LN_V_IP)
#define CDN_MATH_LN_V_IP_USE_BUILTIN
#define CDN_MATH_LN_V_IP cdn_math_ln_v_ip_builtin

static ValueType *cdn_math_ln_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_ln_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LN_V_IP */

#if defined(CDN_MATH_SIGN_V_REQUIRED) && !defined(CDN_MATH_SIGN_V)
#define CDN_MATH_SIGN_V_USE_BUILTIN
#define CDN_MATH_SIGN_V cdn_math_sign_v_builtin

static ValueType *cdn_math_sign_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_sign_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SIGN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SIGN_V */

#if defined(CDN_MATH_SIGN_V_IP_REQUIRED) && !defined(CDN_MATH_SIGN_V_IP)
#define CDN_MATH_SIGN_V_IP_USE_BUILTIN
#define CDN_MATH_SIGN_V_IP cdn_math_sign_v_ip_builtin

static ValueType *cdn_math_sign_v_ip_builtin (ValueType *x0, uint32_t l);

static ValueType *cdn_math_sign_v_ip_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SIGN (x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SIGN_V_IP */

#if defined(CDN_MATH_ATAN2_V_M_M_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_M)
#define CDN_MATH_ATAN2_V_M_M_USE_BUILTIN
#define CDN_MATH_ATAN2_V_M_M cdn_math_atan2_v_m_m_builtin

static ValueType *cdn_math_atan2_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_atan2_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ATAN2 (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_M_M */

#if defined(CDN_MATH_ATAN2_V_M_1_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_1)
#define CDN_MATH_ATAN2_V_M_1_USE_BUILTIN
#define CDN_MATH_ATAN2_V_M_1 cdn_math_atan2_v_m_1_builtin

static ValueType *cdn_math_atan2_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_atan2_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ATAN2 (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_M_1 */

#if defined(CDN_MATH_ATAN2_V_1_M_REQUIRED) && !defined(CDN_MATH_ATAN2_V_1_M)
#define CDN_MATH_ATAN2_V_1_M_USE_BUILTIN
#define CDN_MATH_ATAN2_V_1_M cdn_math_atan2_v_1_m_builtin

static ValueType *cdn_math_atan2_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_atan2_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_ATAN2 (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_1_M */

#if defined(CDN_MATH_ATAN2_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_M_IP)
#define CDN_MATH_ATAN2_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_ATAN2_V_M_M_IP cdn_math_atan2_v_m_m_ip_builtin

static ValueType *cdn_math_atan2_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_atan2_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ATAN2 (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_M_M_IP */

#if defined(CDN_MATH_ATAN2_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_ATAN2_V_M_1_IP)
#define CDN_MATH_ATAN2_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_ATAN2_V_M_1_IP cdn_math_atan2_v_m_1_ip_builtin

static ValueType *cdn_math_atan2_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_atan2_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_ATAN2 (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_M_1_IP */

#if defined(CDN_MATH_ATAN2_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_ATAN2_V_1_M_IP)
#define CDN_MATH_ATAN2_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_ATAN2_V_1_M_IP cdn_math_atan2_v_1_m_ip_builtin

static ValueType *cdn_math_atan2_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_atan2_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_ATAN2 (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_ATAN2_V_1_M_IP */

#if defined(CDN_MATH_POW_V_M_M_REQUIRED) && !defined(CDN_MATH_POW_V_M_M)
#define CDN_MATH_POW_V_M_M_USE_BUILTIN
#define CDN_MATH_POW_V_M_M cdn_math_pow_v_m_m_builtin

static ValueType *cdn_math_pow_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_pow_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_POW (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_M_M */

#if defined(CDN_MATH_POW_V_M_1_REQUIRED) && !defined(CDN_MATH_POW_V_M_1)
#define CDN_MATH_POW_V_M_1_USE_BUILTIN
#define CDN_MATH_POW_V_M_1 cdn_math_pow_v_m_1_builtin

static ValueType *cdn_math_pow_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_pow_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_POW (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_M_1 */

#if defined(CDN_MATH_POW_V_1_M_REQUIRED) && !defined(CDN_MATH_POW_V_1_M)
#define CDN_MATH_POW_V_1_M_USE_BUILTIN
#define CDN_MATH_POW_V_1_M cdn_math_pow_v_1_m_builtin

static ValueType *cdn_math_pow_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_pow_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_POW (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_1_M */

#if defined(CDN_MATH_POW_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_POW_V_M_M_IP)
#define CDN_MATH_POW_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_POW_V_M_M_IP cdn_math_pow_v_m_m_ip_builtin

static ValueType *cdn_math_pow_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_pow_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_POW (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_M_M_IP */

#if defined(CDN_MATH_POW_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_POW_V_M_1_IP)
#define CDN_MATH_POW_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_POW_V_M_1_IP cdn_math_pow_v_m_1_ip_builtin

static ValueType *cdn_math_pow_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_pow_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_POW (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_M_1_IP */

#if defined(CDN_MATH_POW_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_POW_V_1_M_IP)
#define CDN_MATH_POW_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_POW_V_1_M_IP cdn_math_pow_v_1_m_ip_builtin

static ValueType *cdn_math_pow_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_pow_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_POW (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_POW_V_1_M_IP */

#if defined(CDN_MATH_HYPOT_V_M_M_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_M)
#define CDN_MATH_HYPOT_V_M_M_USE_BUILTIN
#define CDN_MATH_HYPOT_V_M_M cdn_math_hypot_v_m_m_builtin

static ValueType *cdn_math_hypot_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_hypot_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_HYPOT (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_M_M */

#if defined(CDN_MATH_HYPOT_V_M_1_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_1)
#define CDN_MATH_HYPOT_V_M_1_USE_BUILTIN
#define CDN_MATH_HYPOT_V_M_1 cdn_math_hypot_v_m_1_builtin

static ValueType *cdn_math_hypot_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_hypot_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_HYPOT (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_M_1 */

#if defined(CDN_MATH_HYPOT_V_1_M_REQUIRED) && !defined(CDN_MATH_HYPOT_V_1_M)
#define CDN_MATH_HYPOT_V_1_M_USE_BUILTIN
#define CDN_MATH_HYPOT_V_1_M cdn_math_hypot_v_1_m_builtin

static ValueType *cdn_math_hypot_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_hypot_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_HYPOT (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_1_M */

#if defined(CDN_MATH_HYPOT_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_M_IP)
#define CDN_MATH_HYPOT_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_HYPOT_V_M_M_IP cdn_math_hypot_v_m_m_ip_builtin

static ValueType *cdn_math_hypot_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_hypot_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_HYPOT (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_M_M_IP */

#if defined(CDN_MATH_HYPOT_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_HYPOT_V_M_1_IP)
#define CDN_MATH_HYPOT_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_HYPOT_V_M_1_IP cdn_math_hypot_v_m_1_ip_builtin

static ValueType *cdn_math_hypot_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_hypot_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_HYPOT (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_M_1_IP */

#if defined(CDN_MATH_HYPOT_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_HYPOT_V_1_M_IP)
#define CDN_MATH_HYPOT_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_HYPOT_V_1_M_IP cdn_math_hypot_v_1_m_ip_builtin

static ValueType *cdn_math_hypot_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_hypot_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_HYPOT (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_HYPOT_V_1_M_IP */

#if defined(CDN_MATH_MAX_V_M_M_REQUIRED) && !defined(CDN_MATH_MAX_V_M_M)
#define CDN_MATH_MAX_V_M_M_USE_BUILTIN
#define CDN_MATH_MAX_V_M_M cdn_math_max_v_m_m_builtin

static ValueType *cdn_math_max_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_max_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MAX (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_M_M */

#if defined(CDN_MATH_MAX_V_M_1_REQUIRED) && !defined(CDN_MATH_MAX_V_M_1)
#define CDN_MATH_MAX_V_M_1_USE_BUILTIN
#define CDN_MATH_MAX_V_M_1 cdn_math_max_v_m_1_builtin

static ValueType *cdn_math_max_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_max_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MAX (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_M_1 */

#if defined(CDN_MATH_MAX_V_1_M_REQUIRED) && !defined(CDN_MATH_MAX_V_1_M)
#define CDN_MATH_MAX_V_1_M_USE_BUILTIN
#define CDN_MATH_MAX_V_1_M cdn_math_max_v_1_m_builtin

static ValueType *cdn_math_max_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_max_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MAX (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_1_M */

#if defined(CDN_MATH_MAX_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_MAX_V_M_M_IP)
#define CDN_MATH_MAX_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_MAX_V_M_M_IP cdn_math_max_v_m_m_ip_builtin

static ValueType *cdn_math_max_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_max_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MAX (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_M_M_IP */

#if defined(CDN_MATH_MAX_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_MAX_V_M_1_IP)
#define CDN_MATH_MAX_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_MAX_V_M_1_IP cdn_math_max_v_m_1_ip_builtin

static ValueType *cdn_math_max_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_max_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MAX (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_M_1_IP */

#if defined(CDN_MATH_MAX_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_MAX_V_1_M_IP)
#define CDN_MATH_MAX_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_MAX_V_1_M_IP cdn_math_max_v_1_m_ip_builtin

static ValueType *cdn_math_max_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_max_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_MAX (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V_1_M_IP */

#if defined(CDN_MATH_MIN_V_M_M_REQUIRED) && !defined(CDN_MATH_MIN_V_M_M)
#define CDN_MATH_MIN_V_M_M_USE_BUILTIN
#define CDN_MATH_MIN_V_M_M cdn_math_min_v_m_m_builtin

static ValueType *cdn_math_min_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_min_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MIN (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_M_M */

#if defined(CDN_MATH_MIN_V_M_1_REQUIRED) && !defined(CDN_MATH_MIN_V_M_1)
#define CDN_MATH_MIN_V_M_1_USE_BUILTIN
#define CDN_MATH_MIN_V_M_1 cdn_math_min_v_m_1_builtin

static ValueType *cdn_math_min_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_min_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MIN (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_M_1 */

#if defined(CDN_MATH_MIN_V_1_M_REQUIRED) && !defined(CDN_MATH_MIN_V_1_M)
#define CDN_MATH_MIN_V_1_M_USE_BUILTIN
#define CDN_MATH_MIN_V_1_M cdn_math_min_v_1_m_builtin

static ValueType *cdn_math_min_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_min_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MIN (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_1_M */

#if defined(CDN_MATH_MIN_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_MIN_V_M_M_IP)
#define CDN_MATH_MIN_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_MIN_V_M_M_IP cdn_math_min_v_m_m_ip_builtin

static ValueType *cdn_math_min_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_min_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MIN (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_M_M_IP */

#if defined(CDN_MATH_MIN_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_MIN_V_M_1_IP)
#define CDN_MATH_MIN_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_MIN_V_M_1_IP cdn_math_min_v_m_1_ip_builtin

static ValueType *cdn_math_min_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_min_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MIN (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_M_1_IP */

#if defined(CDN_MATH_MIN_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_MIN_V_1_M_IP)
#define CDN_MATH_MIN_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_MIN_V_1_M_IP cdn_math_min_v_1_m_ip_builtin

static ValueType *cdn_math_min_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_min_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_MIN (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V_1_M_IP */

#if defined(CDN_MATH_MODULO_V_M_M_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_M)
#define CDN_MATH_MODULO_V_M_M_USE_BUILTIN
#define CDN_MATH_MODULO_V_M_M cdn_math_modulo_v_m_m_builtin

static ValueType *cdn_math_modulo_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_modulo_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MODULO (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_M_M */

#if defined(CDN_MATH_MODULO_V_M_1_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_1)
#define CDN_MATH_MODULO_V_M_1_USE_BUILTIN
#define CDN_MATH_MODULO_V_M_1 cdn_math_modulo_v_m_1_builtin

static ValueType *cdn_math_modulo_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_modulo_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MODULO (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_M_1 */

#if defined(CDN_MATH_MODULO_V_1_M_REQUIRED) && !defined(CDN_MATH_MODULO_V_1_M)
#define CDN_MATH_MODULO_V_1_M_USE_BUILTIN
#define CDN_MATH_MODULO_V_1_M cdn_math_modulo_v_1_m_builtin

static ValueType *cdn_math_modulo_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_modulo_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_MODULO (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_1_M */

#if defined(CDN_MATH_MODULO_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_M_IP)
#define CDN_MATH_MODULO_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_MODULO_V_M_M_IP cdn_math_modulo_v_m_m_ip_builtin

static ValueType *cdn_math_modulo_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_modulo_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MODULO (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_M_M_IP */

#if defined(CDN_MATH_MODULO_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_MODULO_V_M_1_IP)
#define CDN_MATH_MODULO_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_MODULO_V_M_1_IP cdn_math_modulo_v_m_1_ip_builtin

static ValueType *cdn_math_modulo_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_modulo_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_MODULO (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_M_1_IP */

#if defined(CDN_MATH_MODULO_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_MODULO_V_1_M_IP)
#define CDN_MATH_MODULO_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_MODULO_V_1_M_IP cdn_math_modulo_v_1_m_ip_builtin

static ValueType *cdn_math_modulo_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_modulo_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_MODULO (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MODULO_V_1_M_IP */

#if defined(CDN_MATH_CSIGN_V_M_M_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_M)
#define CDN_MATH_CSIGN_V_M_M_USE_BUILTIN
#define CDN_MATH_CSIGN_V_M_M cdn_math_csign_v_m_m_builtin

static ValueType *cdn_math_csign_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_csign_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CSIGN (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_M_M */

#if defined(CDN_MATH_CSIGN_V_M_1_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_1)
#define CDN_MATH_CSIGN_V_M_1_USE_BUILTIN
#define CDN_MATH_CSIGN_V_M_1 cdn_math_csign_v_m_1_builtin

static ValueType *cdn_math_csign_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_csign_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CSIGN (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_M_1 */

#if defined(CDN_MATH_CSIGN_V_1_M_REQUIRED) && !defined(CDN_MATH_CSIGN_V_1_M)
#define CDN_MATH_CSIGN_V_1_M_USE_BUILTIN
#define CDN_MATH_CSIGN_V_1_M cdn_math_csign_v_1_m_builtin

static ValueType *cdn_math_csign_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_csign_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CSIGN (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_1_M */

#if defined(CDN_MATH_CSIGN_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_M_IP)
#define CDN_MATH_CSIGN_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_CSIGN_V_M_M_IP cdn_math_csign_v_m_m_ip_builtin

static ValueType *cdn_math_csign_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_csign_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CSIGN (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_M_M_IP */

#if defined(CDN_MATH_CSIGN_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_CSIGN_V_M_1_IP)
#define CDN_MATH_CSIGN_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_CSIGN_V_M_1_IP cdn_math_csign_v_m_1_ip_builtin

static ValueType *cdn_math_csign_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_csign_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CSIGN (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_M_1_IP */

#if defined(CDN_MATH_CSIGN_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_CSIGN_V_1_M_IP)
#define CDN_MATH_CSIGN_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_CSIGN_V_1_M_IP cdn_math_csign_v_1_m_ip_builtin

static ValueType *cdn_math_csign_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_csign_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_CSIGN (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CSIGN_V_1_M_IP */

#if defined(CDN_MATH_SUM_V_M_M_REQUIRED) && !defined(CDN_MATH_SUM_V_M_M)
#define CDN_MATH_SUM_V_M_M_USE_BUILTIN
#define CDN_MATH_SUM_V_M_M cdn_math_sum_v_m_m_builtin

static ValueType *cdn_math_sum_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sum_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SUM (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_M_M */

#if defined(CDN_MATH_SUM_V_M_1_REQUIRED) && !defined(CDN_MATH_SUM_V_M_1)
#define CDN_MATH_SUM_V_M_1_USE_BUILTIN
#define CDN_MATH_SUM_V_M_1 cdn_math_sum_v_m_1_builtin

static ValueType *cdn_math_sum_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_sum_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SUM (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_M_1 */

#if defined(CDN_MATH_SUM_V_1_M_REQUIRED) && !defined(CDN_MATH_SUM_V_1_M)
#define CDN_MATH_SUM_V_1_M_USE_BUILTIN
#define CDN_MATH_SUM_V_1_M cdn_math_sum_v_1_m_builtin

static ValueType *cdn_math_sum_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sum_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SUM (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_1_M */

#if defined(CDN_MATH_SUM_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_SUM_V_M_M_IP)
#define CDN_MATH_SUM_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_SUM_V_M_M_IP cdn_math_sum_v_m_m_ip_builtin

static ValueType *cdn_math_sum_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sum_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SUM (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_M_M_IP */

#if defined(CDN_MATH_SUM_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_SUM_V_M_1_IP)
#define CDN_MATH_SUM_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_SUM_V_M_1_IP cdn_math_sum_v_m_1_ip_builtin

static ValueType *cdn_math_sum_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_sum_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SUM (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_M_1_IP */

#if defined(CDN_MATH_SUM_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_SUM_V_1_M_IP)
#define CDN_MATH_SUM_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_SUM_V_1_M_IP cdn_math_sum_v_1_m_ip_builtin

static ValueType *cdn_math_sum_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sum_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_SUM (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V_1_M_IP */

#if defined(CDN_MATH_PRODUCT_V_M_M_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_M)
#define CDN_MATH_PRODUCT_V_M_M_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_M_M cdn_math_product_v_m_m_builtin

static ValueType *cdn_math_product_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_product_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_PRODUCT (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_M_M */

#if defined(CDN_MATH_PRODUCT_V_M_1_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_1)
#define CDN_MATH_PRODUCT_V_M_1_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_M_1 cdn_math_product_v_m_1_builtin

static ValueType *cdn_math_product_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_product_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_PRODUCT (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_M_1 */

#if defined(CDN_MATH_PRODUCT_V_1_M_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_1_M)
#define CDN_MATH_PRODUCT_V_1_M_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_1_M cdn_math_product_v_1_m_builtin

static ValueType *cdn_math_product_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_product_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_PRODUCT (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_1_M */

#if defined(CDN_MATH_PRODUCT_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_M_IP)
#define CDN_MATH_PRODUCT_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_M_M_IP cdn_math_product_v_m_m_ip_builtin

static ValueType *cdn_math_product_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_product_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_PRODUCT (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_M_M_IP */

#if defined(CDN_MATH_PRODUCT_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_M_1_IP)
#define CDN_MATH_PRODUCT_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_M_1_IP cdn_math_product_v_m_1_ip_builtin

static ValueType *cdn_math_product_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_product_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_PRODUCT (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_M_1_IP */

#if defined(CDN_MATH_PRODUCT_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_PRODUCT_V_1_M_IP)
#define CDN_MATH_PRODUCT_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_PRODUCT_V_1_M_IP cdn_math_product_v_1_m_ip_builtin

static ValueType *cdn_math_product_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_product_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_PRODUCT (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V_1_M_IP */

#if defined(CDN_MATH_SQSUM_V_M_M_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_M)
#define CDN_MATH_SQSUM_V_M_M_USE_BUILTIN
#define CDN_MATH_SQSUM_V_M_M cdn_math_sqsum_v_m_m_builtin

static ValueType *cdn_math_sqsum_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SQSUM (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_M_M */

#if defined(CDN_MATH_SQSUM_V_M_1_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_1)
#define CDN_MATH_SQSUM_V_M_1_USE_BUILTIN
#define CDN_MATH_SQSUM_V_M_1 cdn_math_sqsum_v_m_1_builtin

static ValueType *cdn_math_sqsum_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SQSUM (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_M_1 */

#if defined(CDN_MATH_SQSUM_V_1_M_REQUIRED) && !defined(CDN_MATH_SQSUM_V_1_M)
#define CDN_MATH_SQSUM_V_1_M_USE_BUILTIN
#define CDN_MATH_SQSUM_V_1_M cdn_math_sqsum_v_1_m_builtin

static ValueType *cdn_math_sqsum_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_SQSUM (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_1_M */

#if defined(CDN_MATH_SQSUM_V_M_M_IP_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_M_IP)
#define CDN_MATH_SQSUM_V_M_M_IP_USE_BUILTIN
#define CDN_MATH_SQSUM_V_M_M_IP cdn_math_sqsum_v_m_m_ip_builtin

static ValueType *cdn_math_sqsum_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_m_m_ip_builtin (ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SQSUM (x0[i], x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_M_M_IP */

#if defined(CDN_MATH_SQSUM_V_M_1_IP_REQUIRED) && !defined(CDN_MATH_SQSUM_V_M_1_IP)
#define CDN_MATH_SQSUM_V_M_1_IP_USE_BUILTIN
#define CDN_MATH_SQSUM_V_M_1_IP cdn_math_sqsum_v_m_1_ip_builtin

static ValueType *cdn_math_sqsum_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_m_1_ip_builtin (ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_SQSUM (x0[i], x1);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_M_1_IP */

#if defined(CDN_MATH_SQSUM_V_1_M_IP_REQUIRED) && !defined(CDN_MATH_SQSUM_V_1_M_IP)
#define CDN_MATH_SQSUM_V_1_M_IP_USE_BUILTIN
#define CDN_MATH_SQSUM_V_1_M_IP cdn_math_sqsum_v_1_m_ip_builtin

static ValueType *cdn_math_sqsum_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_sqsum_v_1_m_ip_builtin (ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_SQSUM (x0, x1[i]);
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V_1_M_IP */

#if defined(CDN_MATH_LERP_V_M_M_M_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_M)
#define CDN_MATH_LERP_V_M_M_M_USE_BUILTIN
#define CDN_MATH_LERP_V_M_M_M cdn_math_lerp_v_m_m_m_builtin

static ValueType *cdn_math_lerp_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_M_M */

#if defined(CDN_MATH_LERP_V_M_M_1_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_1)
#define CDN_MATH_LERP_V_M_M_1_USE_BUILTIN
#define CDN_MATH_LERP_V_M_M_1 cdn_math_lerp_v_m_m_1_builtin

static ValueType *cdn_math_lerp_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_M_1 */

#if defined(CDN_MATH_LERP_V_M_1_M_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_M)
#define CDN_MATH_LERP_V_M_1_M_USE_BUILTIN
#define CDN_MATH_LERP_V_M_1_M cdn_math_lerp_v_m_1_m_builtin

static ValueType *cdn_math_lerp_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_1_M */

#if defined(CDN_MATH_LERP_V_M_1_1_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_1)
#define CDN_MATH_LERP_V_M_1_1_USE_BUILTIN
#define CDN_MATH_LERP_V_M_1_1 cdn_math_lerp_v_m_1_1_builtin

static ValueType *cdn_math_lerp_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_1_1 */

#if defined(CDN_MATH_LERP_V_1_M_M_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_M)
#define CDN_MATH_LERP_V_1_M_M_USE_BUILTIN
#define CDN_MATH_LERP_V_1_M_M cdn_math_lerp_v_1_m_m_builtin

static ValueType *cdn_math_lerp_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_M_M */

#if defined(CDN_MATH_LERP_V_1_M_1_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_1)
#define CDN_MATH_LERP_V_1_M_1_USE_BUILTIN
#define CDN_MATH_LERP_V_1_M_1 cdn_math_lerp_v_1_m_1_builtin

static ValueType *cdn_math_lerp_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_M_1 */

#if defined(CDN_MATH_LERP_V_1_1_M_REQUIRED) && !defined(CDN_MATH_LERP_V_1_1_M)
#define CDN_MATH_LERP_V_1_1_M_USE_BUILTIN
#define CDN_MATH_LERP_V_1_1_M cdn_math_lerp_v_1_1_m_builtin

static ValueType *cdn_math_lerp_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_LERP (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_1_M */

#if defined(CDN_MATH_LERP_V_M_M_M_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_M_IP)
#define CDN_MATH_LERP_V_M_M_M_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_M_M_M_IP cdn_math_lerp_v_m_m_m_ip_builtin

static ValueType *cdn_math_lerp_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LERP (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_M_M_IP */

#if defined(CDN_MATH_LERP_V_M_M_1_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_M_M_1_IP)
#define CDN_MATH_LERP_V_M_M_1_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_M_M_1_IP cdn_math_lerp_v_m_m_1_ip_builtin

static ValueType *cdn_math_lerp_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LERP (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_M_1_IP */

#if defined(CDN_MATH_LERP_V_M_1_M_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_M_IP)
#define CDN_MATH_LERP_V_M_1_M_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_M_1_M_IP cdn_math_lerp_v_m_1_m_ip_builtin

static ValueType *cdn_math_lerp_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LERP (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_1_M_IP */

#if defined(CDN_MATH_LERP_V_M_1_1_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_M_1_1_IP)
#define CDN_MATH_LERP_V_M_1_1_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_M_1_1_IP cdn_math_lerp_v_m_1_1_ip_builtin

static ValueType *cdn_math_lerp_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_LERP (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_M_1_1_IP */

#if defined(CDN_MATH_LERP_V_1_M_M_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_M_IP)
#define CDN_MATH_LERP_V_1_M_M_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_1_M_M_IP cdn_math_lerp_v_1_m_m_ip_builtin

static ValueType *cdn_math_lerp_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_LERP (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_M_M_IP */

#if defined(CDN_MATH_LERP_V_1_M_1_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_1_M_1_IP)
#define CDN_MATH_LERP_V_1_M_1_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_1_M_1_IP cdn_math_lerp_v_1_m_1_ip_builtin

static ValueType *cdn_math_lerp_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_LERP (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_M_1_IP */

#if defined(CDN_MATH_LERP_V_1_1_M_IP_REQUIRED) && !defined(CDN_MATH_LERP_V_1_1_M_IP)
#define CDN_MATH_LERP_V_1_1_M_IP_USE_BUILTIN
#define CDN_MATH_LERP_V_1_1_M_IP cdn_math_lerp_v_1_1_m_ip_builtin

static ValueType *cdn_math_lerp_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_lerp_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x2[i] = CDN_MATH_LERP (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_LERP_V_1_1_M_IP */

#if defined(CDN_MATH_CLIP_V_M_M_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_M)
#define CDN_MATH_CLIP_V_M_M_M_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_M_M cdn_math_clip_v_m_m_m_builtin

static ValueType *cdn_math_clip_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_M_M */

#if defined(CDN_MATH_CLIP_V_M_M_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_1)
#define CDN_MATH_CLIP_V_M_M_1_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_M_1 cdn_math_clip_v_m_m_1_builtin

static ValueType *cdn_math_clip_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_M_1 */

#if defined(CDN_MATH_CLIP_V_M_1_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_M)
#define CDN_MATH_CLIP_V_M_1_M_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_1_M cdn_math_clip_v_m_1_m_builtin

static ValueType *cdn_math_clip_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_1_M */

#if defined(CDN_MATH_CLIP_V_M_1_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_1)
#define CDN_MATH_CLIP_V_M_1_1_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_1_1 cdn_math_clip_v_m_1_1_builtin

static ValueType *cdn_math_clip_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_1_1 */

#if defined(CDN_MATH_CLIP_V_1_M_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_M)
#define CDN_MATH_CLIP_V_1_M_M_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_M_M cdn_math_clip_v_1_m_m_builtin

static ValueType *cdn_math_clip_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_M_M */

#if defined(CDN_MATH_CLIP_V_1_M_1_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_1)
#define CDN_MATH_CLIP_V_1_M_1_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_M_1 cdn_math_clip_v_1_m_1_builtin

static ValueType *cdn_math_clip_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_M_1 */

#if defined(CDN_MATH_CLIP_V_1_1_M_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_1_M)
#define CDN_MATH_CLIP_V_1_1_M_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_1_M cdn_math_clip_v_1_1_m_builtin

static ValueType *cdn_math_clip_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CLIP (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_1_M */

#if defined(CDN_MATH_CLIP_V_M_M_M_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_M_IP)
#define CDN_MATH_CLIP_V_M_M_M_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_M_M_IP cdn_math_clip_v_m_m_m_ip_builtin

static ValueType *cdn_math_clip_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CLIP (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_M_M_IP */

#if defined(CDN_MATH_CLIP_V_M_M_1_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_M_1_IP)
#define CDN_MATH_CLIP_V_M_M_1_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_M_1_IP cdn_math_clip_v_m_m_1_ip_builtin

static ValueType *cdn_math_clip_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CLIP (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_M_1_IP */

#if defined(CDN_MATH_CLIP_V_M_1_M_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_M_IP)
#define CDN_MATH_CLIP_V_M_1_M_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_1_M_IP cdn_math_clip_v_m_1_m_ip_builtin

static ValueType *cdn_math_clip_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CLIP (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_1_M_IP */

#if defined(CDN_MATH_CLIP_V_M_1_1_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_M_1_1_IP)
#define CDN_MATH_CLIP_V_M_1_1_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_M_1_1_IP cdn_math_clip_v_m_1_1_ip_builtin

static ValueType *cdn_math_clip_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CLIP (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_M_1_1_IP */

#if defined(CDN_MATH_CLIP_V_1_M_M_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_M_IP)
#define CDN_MATH_CLIP_V_1_M_M_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_M_M_IP cdn_math_clip_v_1_m_m_ip_builtin

static ValueType *cdn_math_clip_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_CLIP (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_M_M_IP */

#if defined(CDN_MATH_CLIP_V_1_M_1_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_M_1_IP)
#define CDN_MATH_CLIP_V_1_M_1_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_M_1_IP cdn_math_clip_v_1_m_1_ip_builtin

static ValueType *cdn_math_clip_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_CLIP (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_M_1_IP */

#if defined(CDN_MATH_CLIP_V_1_1_M_IP_REQUIRED) && !defined(CDN_MATH_CLIP_V_1_1_M_IP)
#define CDN_MATH_CLIP_V_1_1_M_IP_USE_BUILTIN
#define CDN_MATH_CLIP_V_1_1_M_IP cdn_math_clip_v_1_1_m_ip_builtin

static ValueType *cdn_math_clip_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_clip_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x2[i] = CDN_MATH_CLIP (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CLIP_V_1_1_M_IP */

#if defined(CDN_MATH_CYCLE_V_M_M_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_M)
#define CDN_MATH_CYCLE_V_M_M_M_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_M_M cdn_math_cycle_v_m_m_m_builtin

static ValueType *cdn_math_cycle_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_M_M */

#if defined(CDN_MATH_CYCLE_V_M_M_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_1)
#define CDN_MATH_CYCLE_V_M_M_1_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_M_1 cdn_math_cycle_v_m_m_1_builtin

static ValueType *cdn_math_cycle_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_M_1 */

#if defined(CDN_MATH_CYCLE_V_M_1_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_M)
#define CDN_MATH_CYCLE_V_M_1_M_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_1_M cdn_math_cycle_v_m_1_m_builtin

static ValueType *cdn_math_cycle_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_1_m_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_1_M */

#if defined(CDN_MATH_CYCLE_V_M_1_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_1)
#define CDN_MATH_CYCLE_V_M_1_1_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_1_1 cdn_math_cycle_v_m_1_1_builtin

static ValueType *cdn_math_cycle_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_1_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_1_1 */

#if defined(CDN_MATH_CYCLE_V_1_M_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_M)
#define CDN_MATH_CYCLE_V_1_M_M_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_M_M cdn_math_cycle_v_1_m_m_builtin

static ValueType *cdn_math_cycle_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_m_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_M_M */

#if defined(CDN_MATH_CYCLE_V_1_M_1_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_1)
#define CDN_MATH_CYCLE_V_1_M_1_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_M_1 cdn_math_cycle_v_1_m_1_builtin

static ValueType *cdn_math_cycle_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_m_1_builtin (ValueType *ret, ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_M_1 */

#if defined(CDN_MATH_CYCLE_V_1_1_M_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_1_M)
#define CDN_MATH_CYCLE_V_1_1_M_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_1_M cdn_math_cycle_v_1_1_m_builtin

static ValueType *cdn_math_cycle_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_1_m_builtin (ValueType *ret, ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = CDN_MATH_CYCLE (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_1_M */

#if defined(CDN_MATH_CYCLE_V_M_M_M_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_M_IP)
#define CDN_MATH_CYCLE_V_M_M_M_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_M_M_IP cdn_math_cycle_v_m_m_m_ip_builtin

static ValueType *cdn_math_cycle_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_m_m_ip_builtin (ValueType *x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CYCLE (x0[i], x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_M_M_IP */

#if defined(CDN_MATH_CYCLE_V_M_M_1_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_M_1_IP)
#define CDN_MATH_CYCLE_V_M_M_1_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_M_1_IP cdn_math_cycle_v_m_m_1_ip_builtin

static ValueType *cdn_math_cycle_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_m_1_ip_builtin (ValueType *x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CYCLE (x0[i], x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_M_1_IP */

#if defined(CDN_MATH_CYCLE_V_M_1_M_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_M_IP)
#define CDN_MATH_CYCLE_V_M_1_M_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_1_M_IP cdn_math_cycle_v_m_1_m_ip_builtin

static ValueType *cdn_math_cycle_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_1_m_ip_builtin (ValueType *x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CYCLE (x0[i], x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_1_M_IP */

#if defined(CDN_MATH_CYCLE_V_M_1_1_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_M_1_1_IP)
#define CDN_MATH_CYCLE_V_M_1_1_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_M_1_1_IP cdn_math_cycle_v_m_1_1_ip_builtin

static ValueType *cdn_math_cycle_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_m_1_1_ip_builtin (ValueType *x0, ValueType x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x0[i] = CDN_MATH_CYCLE (x0[i], x1, x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_M_1_1_IP */

#if defined(CDN_MATH_CYCLE_V_1_M_M_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_M_IP)
#define CDN_MATH_CYCLE_V_1_M_M_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_M_M_IP cdn_math_cycle_v_1_m_m_ip_builtin

static ValueType *cdn_math_cycle_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_m_m_ip_builtin (ValueType x0, ValueType *x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_CYCLE (x0, x1[i], x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_M_M_IP */

#if defined(CDN_MATH_CYCLE_V_1_M_1_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_M_1_IP)
#define CDN_MATH_CYCLE_V_1_M_1_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_M_1_IP cdn_math_cycle_v_1_m_1_ip_builtin

static ValueType *cdn_math_cycle_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_m_1_ip_builtin (ValueType x0, ValueType *x1, ValueType x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x1[i] = CDN_MATH_CYCLE (x0, x1[i], x2);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_M_1_IP */

#if defined(CDN_MATH_CYCLE_V_1_1_M_IP_REQUIRED) && !defined(CDN_MATH_CYCLE_V_1_1_M_IP)
#define CDN_MATH_CYCLE_V_1_1_M_IP_USE_BUILTIN
#define CDN_MATH_CYCLE_V_1_1_M_IP cdn_math_cycle_v_1_1_m_ip_builtin

static ValueType *cdn_math_cycle_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l);

static ValueType *cdn_math_cycle_v_1_1_m_ip_builtin (ValueType x0, ValueType x1, ValueType *x2, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		x2[i] = CDN_MATH_CYCLE (x0, x1, x2[i]);
	}

	return ret;
}
#endif /* CDN_MATH_CYCLE_V_1_1_M_IP */

#if defined(CDN_MATH_UMINUS_V_REQUIRED) && !defined(CDN_MATH_UMINUS_V)
#define CDN_MATH_UMINUS_V_USE_BUILTIN
#define CDN_MATH_UMINUS_V cdn_math_uminus_v_builtin

static ValueType *cdn_math_uminus_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_uminus_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = -x0[i];
	}

	return ret;
}
#endif /* CDN_MATH_UMINUS_V */

#if defined(CDN_MATH_NEGATE_V_REQUIRED) && !defined(CDN_MATH_NEGATE_V)
#define CDN_MATH_NEGATE_V_USE_BUILTIN
#define CDN_MATH_NEGATE_V cdn_math_negate_v_builtin

static ValueType *cdn_math_negate_v_builtin (ValueType *ret, ValueType *x0, uint32_t l);

static ValueType *cdn_math_negate_v_builtin (ValueType *ret, ValueType *x0, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = !x0[i];
	}

	return ret;
}
#endif /* CDN_MATH_NEGATE_V */

#if defined(CDN_MATH_MINUS_V_M_M_REQUIRED) && !defined(CDN_MATH_MINUS_V_M_M)
#define CDN_MATH_MINUS_V_M_M_USE_BUILTIN
#define CDN_MATH_MINUS_V_M_M cdn_math_minus_v_m_m_builtin

static ValueType *cdn_math_minus_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_minus_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] - x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_M_M */

#if defined(CDN_MATH_MINUS_V_1_M_REQUIRED) && !defined(CDN_MATH_MINUS_V_1_M)
#define CDN_MATH_MINUS_V_1_M_USE_BUILTIN
#define CDN_MATH_MINUS_V_1_M cdn_math_minus_v_1_m_builtin

static ValueType *cdn_math_minus_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_minus_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 - x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_1_M */

#if defined(CDN_MATH_MINUS_V_M_1_REQUIRED) && !defined(CDN_MATH_MINUS_V_M_1)
#define CDN_MATH_MINUS_V_M_1_USE_BUILTIN
#define CDN_MATH_MINUS_V_M_1 cdn_math_minus_v_m_1_builtin

static ValueType *cdn_math_minus_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_minus_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] - x1;
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_M_1 */

#if defined(CDN_MATH_MINUS_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_MINUS_V_CWISE_M_1)
#define CDN_MATH_MINUS_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_MINUS_V_CWISE_M_1 cdn_math_minus_v_cwise_m_1_builtin

static ValueType *cdn_math_minus_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_minus_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  -  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_CWISE_M_1 */

#if defined(CDN_MATH_MINUS_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_MINUS_V_CWISE_1_M)
#define CDN_MATH_MINUS_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_MINUS_V_CWISE_1_M cdn_math_minus_v_cwise_1_m_builtin

static ValueType *cdn_math_minus_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_minus_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  -  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_CWISE_1_M */

#if defined(CDN_MATH_MINUS_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_MINUS_V_RWISE_M_1)
#define CDN_MATH_MINUS_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_MINUS_V_RWISE_M_1 cdn_math_minus_v_rwise_m_1_builtin

static ValueType *cdn_math_minus_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_minus_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  -  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_RWISE_M_1 */

#if defined(CDN_MATH_MINUS_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_MINUS_V_RWISE_1_M)
#define CDN_MATH_MINUS_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_MINUS_V_RWISE_1_M cdn_math_minus_v_rwise_1_m_builtin

static ValueType *cdn_math_minus_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_minus_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  -  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_MINUS_V_RWISE_1_M */

#if defined(CDN_MATH_PLUS_V_M_M_REQUIRED) && !defined(CDN_MATH_PLUS_V_M_M)
#define CDN_MATH_PLUS_V_M_M_USE_BUILTIN
#define CDN_MATH_PLUS_V_M_M cdn_math_plus_v_m_m_builtin

static ValueType *cdn_math_plus_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_plus_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] + x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_M_M */

#if defined(CDN_MATH_PLUS_V_1_M_REQUIRED) && !defined(CDN_MATH_PLUS_V_1_M)
#define CDN_MATH_PLUS_V_1_M_USE_BUILTIN
#define CDN_MATH_PLUS_V_1_M cdn_math_plus_v_1_m_builtin

static ValueType *cdn_math_plus_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_plus_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 + x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_1_M */

#if defined(CDN_MATH_PLUS_V_M_1_REQUIRED) && !defined(CDN_MATH_PLUS_V_M_1)
#define CDN_MATH_PLUS_V_M_1_USE_BUILTIN
#define CDN_MATH_PLUS_V_M_1 cdn_math_plus_v_m_1_builtin

static ValueType *cdn_math_plus_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_plus_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] + x1;
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_M_1 */

#if defined(CDN_MATH_PLUS_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_PLUS_V_CWISE_M_1)
#define CDN_MATH_PLUS_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_PLUS_V_CWISE_M_1 cdn_math_plus_v_cwise_m_1_builtin

static ValueType *cdn_math_plus_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_plus_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  +  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_CWISE_M_1 */

#if defined(CDN_MATH_PLUS_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_PLUS_V_CWISE_1_M)
#define CDN_MATH_PLUS_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_PLUS_V_CWISE_1_M cdn_math_plus_v_cwise_1_m_builtin

static ValueType *cdn_math_plus_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_plus_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  +  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_CWISE_1_M */

#if defined(CDN_MATH_PLUS_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_PLUS_V_RWISE_M_1)
#define CDN_MATH_PLUS_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_PLUS_V_RWISE_M_1 cdn_math_plus_v_rwise_m_1_builtin

static ValueType *cdn_math_plus_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_plus_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  +  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_RWISE_M_1 */

#if defined(CDN_MATH_PLUS_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_PLUS_V_RWISE_1_M)
#define CDN_MATH_PLUS_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_PLUS_V_RWISE_1_M cdn_math_plus_v_rwise_1_m_builtin

static ValueType *cdn_math_plus_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_plus_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  +  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_PLUS_V_RWISE_1_M */

#if defined(CDN_MATH_EMULTIPLY_V_M_M_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_M_M)
#define CDN_MATH_EMULTIPLY_V_M_M_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_M_M cdn_math_emultiply_v_m_m_builtin

static ValueType *cdn_math_emultiply_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_emultiply_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] * x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_M_M */

#if defined(CDN_MATH_EMULTIPLY_V_1_M_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_1_M)
#define CDN_MATH_EMULTIPLY_V_1_M_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_1_M cdn_math_emultiply_v_1_m_builtin

static ValueType *cdn_math_emultiply_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_emultiply_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 * x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_1_M */

#if defined(CDN_MATH_EMULTIPLY_V_M_1_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_M_1)
#define CDN_MATH_EMULTIPLY_V_M_1_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_M_1 cdn_math_emultiply_v_m_1_builtin

static ValueType *cdn_math_emultiply_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_emultiply_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] * x1;
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_M_1 */

#if defined(CDN_MATH_EMULTIPLY_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_CWISE_M_1)
#define CDN_MATH_EMULTIPLY_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_CWISE_M_1 cdn_math_emultiply_v_cwise_m_1_builtin

static ValueType *cdn_math_emultiply_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_emultiply_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  *  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_CWISE_M_1 */

#if defined(CDN_MATH_EMULTIPLY_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_CWISE_1_M)
#define CDN_MATH_EMULTIPLY_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_CWISE_1_M cdn_math_emultiply_v_cwise_1_m_builtin

static ValueType *cdn_math_emultiply_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_emultiply_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  *  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_CWISE_1_M */

#if defined(CDN_MATH_EMULTIPLY_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_RWISE_M_1)
#define CDN_MATH_EMULTIPLY_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_RWISE_M_1 cdn_math_emultiply_v_rwise_m_1_builtin

static ValueType *cdn_math_emultiply_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_emultiply_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  *  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_RWISE_M_1 */

#if defined(CDN_MATH_EMULTIPLY_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_EMULTIPLY_V_RWISE_1_M)
#define CDN_MATH_EMULTIPLY_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_EMULTIPLY_V_RWISE_1_M cdn_math_emultiply_v_rwise_1_m_builtin

static ValueType *cdn_math_emultiply_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_emultiply_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  *  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EMULTIPLY_V_RWISE_1_M */

#if defined(CDN_MATH_DIVIDE_V_M_M_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_M_M)
#define CDN_MATH_DIVIDE_V_M_M_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_M_M cdn_math_divide_v_m_m_builtin

static ValueType *cdn_math_divide_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_divide_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] / x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_M_M */

#if defined(CDN_MATH_DIVIDE_V_1_M_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_1_M)
#define CDN_MATH_DIVIDE_V_1_M_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_1_M cdn_math_divide_v_1_m_builtin

static ValueType *cdn_math_divide_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_divide_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 / x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_1_M */

#if defined(CDN_MATH_DIVIDE_V_M_1_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_M_1)
#define CDN_MATH_DIVIDE_V_M_1_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_M_1 cdn_math_divide_v_m_1_builtin

static ValueType *cdn_math_divide_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_divide_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] / x1;
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_M_1 */

#if defined(CDN_MATH_DIVIDE_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_CWISE_M_1)
#define CDN_MATH_DIVIDE_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_CWISE_M_1 cdn_math_divide_v_cwise_m_1_builtin

static ValueType *cdn_math_divide_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_divide_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  /  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_CWISE_M_1 */

#if defined(CDN_MATH_DIVIDE_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_CWISE_1_M)
#define CDN_MATH_DIVIDE_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_CWISE_1_M cdn_math_divide_v_cwise_1_m_builtin

static ValueType *cdn_math_divide_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_divide_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  /  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_CWISE_1_M */

#if defined(CDN_MATH_DIVIDE_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_RWISE_M_1)
#define CDN_MATH_DIVIDE_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_RWISE_M_1 cdn_math_divide_v_rwise_m_1_builtin

static ValueType *cdn_math_divide_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_divide_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  /  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_RWISE_M_1 */

#if defined(CDN_MATH_DIVIDE_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_DIVIDE_V_RWISE_1_M)
#define CDN_MATH_DIVIDE_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_DIVIDE_V_RWISE_1_M cdn_math_divide_v_rwise_1_m_builtin

static ValueType *cdn_math_divide_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_divide_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  /  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_DIVIDE_V_RWISE_1_M */

#if defined(CDN_MATH_GREATER_V_M_M_REQUIRED) && !defined(CDN_MATH_GREATER_V_M_M)
#define CDN_MATH_GREATER_V_M_M_USE_BUILTIN
#define CDN_MATH_GREATER_V_M_M cdn_math_greater_v_m_m_builtin

static ValueType *cdn_math_greater_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_greater_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] > x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_M_M */

#if defined(CDN_MATH_GREATER_V_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_V_1_M)
#define CDN_MATH_GREATER_V_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_V_1_M cdn_math_greater_v_1_m_builtin

static ValueType *cdn_math_greater_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_greater_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 > x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_1_M */

#if defined(CDN_MATH_GREATER_V_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_V_M_1)
#define CDN_MATH_GREATER_V_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_V_M_1 cdn_math_greater_v_m_1_builtin

static ValueType *cdn_math_greater_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_greater_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] > x1;
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_M_1 */

#if defined(CDN_MATH_GREATER_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_V_CWISE_M_1)
#define CDN_MATH_GREATER_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_V_CWISE_M_1 cdn_math_greater_v_cwise_m_1_builtin

static ValueType *cdn_math_greater_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  >  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_CWISE_M_1 */

#if defined(CDN_MATH_GREATER_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_V_CWISE_1_M)
#define CDN_MATH_GREATER_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_V_CWISE_1_M cdn_math_greater_v_cwise_1_m_builtin

static ValueType *cdn_math_greater_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  >  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_CWISE_1_M */

#if defined(CDN_MATH_GREATER_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_V_RWISE_M_1)
#define CDN_MATH_GREATER_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_V_RWISE_M_1 cdn_math_greater_v_rwise_m_1_builtin

static ValueType *cdn_math_greater_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  >  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_RWISE_M_1 */

#if defined(CDN_MATH_GREATER_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_V_RWISE_1_M)
#define CDN_MATH_GREATER_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_V_RWISE_1_M cdn_math_greater_v_rwise_1_m_builtin

static ValueType *cdn_math_greater_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  >  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_V_RWISE_1_M */

#if defined(CDN_MATH_LESS_V_M_M_REQUIRED) && !defined(CDN_MATH_LESS_V_M_M)
#define CDN_MATH_LESS_V_M_M_USE_BUILTIN
#define CDN_MATH_LESS_V_M_M cdn_math_less_v_m_m_builtin

static ValueType *cdn_math_less_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_less_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] < x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_M_M */

#if defined(CDN_MATH_LESS_V_1_M_REQUIRED) && !defined(CDN_MATH_LESS_V_1_M)
#define CDN_MATH_LESS_V_1_M_USE_BUILTIN
#define CDN_MATH_LESS_V_1_M cdn_math_less_v_1_m_builtin

static ValueType *cdn_math_less_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_less_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 < x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_1_M */

#if defined(CDN_MATH_LESS_V_M_1_REQUIRED) && !defined(CDN_MATH_LESS_V_M_1)
#define CDN_MATH_LESS_V_M_1_USE_BUILTIN
#define CDN_MATH_LESS_V_M_1 cdn_math_less_v_m_1_builtin

static ValueType *cdn_math_less_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_less_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] < x1;
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_M_1 */

#if defined(CDN_MATH_LESS_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_LESS_V_CWISE_M_1)
#define CDN_MATH_LESS_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_LESS_V_CWISE_M_1 cdn_math_less_v_cwise_m_1_builtin

static ValueType *cdn_math_less_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  <  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_CWISE_M_1 */

#if defined(CDN_MATH_LESS_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_LESS_V_CWISE_1_M)
#define CDN_MATH_LESS_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_LESS_V_CWISE_1_M cdn_math_less_v_cwise_1_m_builtin

static ValueType *cdn_math_less_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  <  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_CWISE_1_M */

#if defined(CDN_MATH_LESS_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_LESS_V_RWISE_M_1)
#define CDN_MATH_LESS_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_LESS_V_RWISE_M_1 cdn_math_less_v_rwise_m_1_builtin

static ValueType *cdn_math_less_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  <  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_RWISE_M_1 */

#if defined(CDN_MATH_LESS_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_LESS_V_RWISE_1_M)
#define CDN_MATH_LESS_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_LESS_V_RWISE_1_M cdn_math_less_v_rwise_1_m_builtin

static ValueType *cdn_math_less_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  <  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_V_RWISE_1_M */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_M_M_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_M_M)
#define CDN_MATH_GREATER_OR_EQUAL_V_M_M_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_M_M cdn_math_greater_or_equal_v_m_m_builtin

static ValueType *cdn_math_greater_or_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_greater_or_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] >= x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_M_M */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_1_M)
#define CDN_MATH_GREATER_OR_EQUAL_V_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_1_M cdn_math_greater_or_equal_v_1_m_builtin

static ValueType *cdn_math_greater_or_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_greater_or_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 >= x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_1_M */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_M_1)
#define CDN_MATH_GREATER_OR_EQUAL_V_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_M_1 cdn_math_greater_or_equal_v_m_1_builtin

static ValueType *cdn_math_greater_or_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_greater_or_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] >= x1;
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_M_1 */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_CWISE_M_1)
#define CDN_MATH_GREATER_OR_EQUAL_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_CWISE_M_1 cdn_math_greater_or_equal_v_cwise_m_1_builtin

static ValueType *cdn_math_greater_or_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_or_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  >=  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_CWISE_M_1 */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_CWISE_1_M)
#define CDN_MATH_GREATER_OR_EQUAL_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_CWISE_1_M cdn_math_greater_or_equal_v_cwise_1_m_builtin

static ValueType *cdn_math_greater_or_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_or_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  >=  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_CWISE_1_M */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_RWISE_M_1)
#define CDN_MATH_GREATER_OR_EQUAL_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_RWISE_M_1 cdn_math_greater_or_equal_v_rwise_m_1_builtin

static ValueType *cdn_math_greater_or_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_or_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  >=  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_RWISE_M_1 */

#if defined(CDN_MATH_GREATER_OR_EQUAL_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_GREATER_OR_EQUAL_V_RWISE_1_M)
#define CDN_MATH_GREATER_OR_EQUAL_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_GREATER_OR_EQUAL_V_RWISE_1_M cdn_math_greater_or_equal_v_rwise_1_m_builtin

static ValueType *cdn_math_greater_or_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_greater_or_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  >=  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_GREATER_OR_EQUAL_V_RWISE_1_M */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_M_M_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_M_M)
#define CDN_MATH_LESS_OR_EQUAL_V_M_M_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_M_M cdn_math_less_or_equal_v_m_m_builtin

static ValueType *cdn_math_less_or_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_less_or_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] <= x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_M_M */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_1_M_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_1_M)
#define CDN_MATH_LESS_OR_EQUAL_V_1_M_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_1_M cdn_math_less_or_equal_v_1_m_builtin

static ValueType *cdn_math_less_or_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_less_or_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 <= x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_1_M */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_M_1_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_M_1)
#define CDN_MATH_LESS_OR_EQUAL_V_M_1_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_M_1 cdn_math_less_or_equal_v_m_1_builtin

static ValueType *cdn_math_less_or_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_less_or_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] <= x1;
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_M_1 */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_CWISE_M_1)
#define CDN_MATH_LESS_OR_EQUAL_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_CWISE_M_1 cdn_math_less_or_equal_v_cwise_m_1_builtin

static ValueType *cdn_math_less_or_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_or_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  <=  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_CWISE_M_1 */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_CWISE_1_M)
#define CDN_MATH_LESS_OR_EQUAL_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_CWISE_1_M cdn_math_less_or_equal_v_cwise_1_m_builtin

static ValueType *cdn_math_less_or_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_or_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  <=  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_CWISE_1_M */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_RWISE_M_1)
#define CDN_MATH_LESS_OR_EQUAL_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_RWISE_M_1 cdn_math_less_or_equal_v_rwise_m_1_builtin

static ValueType *cdn_math_less_or_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_or_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  <=  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_RWISE_M_1 */

#if defined(CDN_MATH_LESS_OR_EQUAL_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_LESS_OR_EQUAL_V_RWISE_1_M)
#define CDN_MATH_LESS_OR_EQUAL_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_LESS_OR_EQUAL_V_RWISE_1_M cdn_math_less_or_equal_v_rwise_1_m_builtin

static ValueType *cdn_math_less_or_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_less_or_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  <=  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_LESS_OR_EQUAL_V_RWISE_1_M */

#if defined(CDN_MATH_EQUAL_V_M_M_REQUIRED) && !defined(CDN_MATH_EQUAL_V_M_M)
#define CDN_MATH_EQUAL_V_M_M_USE_BUILTIN
#define CDN_MATH_EQUAL_V_M_M cdn_math_equal_v_m_m_builtin

static ValueType *cdn_math_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_equal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] == x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_M_M */

#if defined(CDN_MATH_EQUAL_V_1_M_REQUIRED) && !defined(CDN_MATH_EQUAL_V_1_M)
#define CDN_MATH_EQUAL_V_1_M_USE_BUILTIN
#define CDN_MATH_EQUAL_V_1_M cdn_math_equal_v_1_m_builtin

static ValueType *cdn_math_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_equal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 == x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_1_M */

#if defined(CDN_MATH_EQUAL_V_M_1_REQUIRED) && !defined(CDN_MATH_EQUAL_V_M_1)
#define CDN_MATH_EQUAL_V_M_1_USE_BUILTIN
#define CDN_MATH_EQUAL_V_M_1 cdn_math_equal_v_m_1_builtin

static ValueType *cdn_math_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_equal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] == x1;
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_M_1 */

#if defined(CDN_MATH_EQUAL_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_EQUAL_V_CWISE_M_1)
#define CDN_MATH_EQUAL_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_EQUAL_V_CWISE_M_1 cdn_math_equal_v_cwise_m_1_builtin

static ValueType *cdn_math_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_equal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  ==  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_CWISE_M_1 */

#if defined(CDN_MATH_EQUAL_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_EQUAL_V_CWISE_1_M)
#define CDN_MATH_EQUAL_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_EQUAL_V_CWISE_1_M cdn_math_equal_v_cwise_1_m_builtin

static ValueType *cdn_math_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_equal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  ==  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_CWISE_1_M */

#if defined(CDN_MATH_EQUAL_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_EQUAL_V_RWISE_M_1)
#define CDN_MATH_EQUAL_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_EQUAL_V_RWISE_M_1 cdn_math_equal_v_rwise_m_1_builtin

static ValueType *cdn_math_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_equal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  ==  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_RWISE_M_1 */

#if defined(CDN_MATH_EQUAL_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_EQUAL_V_RWISE_1_M)
#define CDN_MATH_EQUAL_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_EQUAL_V_RWISE_1_M cdn_math_equal_v_rwise_1_m_builtin

static ValueType *cdn_math_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_equal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  ==  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_EQUAL_V_RWISE_1_M */

#if defined(CDN_MATH_NEQUAL_V_M_M_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_M_M)
#define CDN_MATH_NEQUAL_V_M_M_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_M_M cdn_math_nequal_v_m_m_builtin

static ValueType *cdn_math_nequal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_nequal_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] != x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_M_M */

#if defined(CDN_MATH_NEQUAL_V_1_M_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_1_M)
#define CDN_MATH_NEQUAL_V_1_M_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_1_M cdn_math_nequal_v_1_m_builtin

static ValueType *cdn_math_nequal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_nequal_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 != x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_1_M */

#if defined(CDN_MATH_NEQUAL_V_M_1_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_M_1)
#define CDN_MATH_NEQUAL_V_M_1_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_M_1 cdn_math_nequal_v_m_1_builtin

static ValueType *cdn_math_nequal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_nequal_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] != x1;
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_M_1 */

#if defined(CDN_MATH_NEQUAL_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_CWISE_M_1)
#define CDN_MATH_NEQUAL_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_CWISE_M_1 cdn_math_nequal_v_cwise_m_1_builtin

static ValueType *cdn_math_nequal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_nequal_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  !=  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_CWISE_M_1 */

#if defined(CDN_MATH_NEQUAL_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_CWISE_1_M)
#define CDN_MATH_NEQUAL_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_CWISE_1_M cdn_math_nequal_v_cwise_1_m_builtin

static ValueType *cdn_math_nequal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_nequal_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  !=  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_CWISE_1_M */

#if defined(CDN_MATH_NEQUAL_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_RWISE_M_1)
#define CDN_MATH_NEQUAL_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_RWISE_M_1 cdn_math_nequal_v_rwise_m_1_builtin

static ValueType *cdn_math_nequal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_nequal_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  !=  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_RWISE_M_1 */

#if defined(CDN_MATH_NEQUAL_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_NEQUAL_V_RWISE_1_M)
#define CDN_MATH_NEQUAL_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_NEQUAL_V_RWISE_1_M cdn_math_nequal_v_rwise_1_m_builtin

static ValueType *cdn_math_nequal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_nequal_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  !=  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_NEQUAL_V_RWISE_1_M */

#if defined(CDN_MATH_OR_V_M_M_REQUIRED) && !defined(CDN_MATH_OR_V_M_M)
#define CDN_MATH_OR_V_M_M_USE_BUILTIN
#define CDN_MATH_OR_V_M_M cdn_math_or_v_m_m_builtin

static ValueType *cdn_math_or_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_or_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] || x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_M_M */

#if defined(CDN_MATH_OR_V_1_M_REQUIRED) && !defined(CDN_MATH_OR_V_1_M)
#define CDN_MATH_OR_V_1_M_USE_BUILTIN
#define CDN_MATH_OR_V_1_M cdn_math_or_v_1_m_builtin

static ValueType *cdn_math_or_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_or_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 || x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_1_M */

#if defined(CDN_MATH_OR_V_M_1_REQUIRED) && !defined(CDN_MATH_OR_V_M_1)
#define CDN_MATH_OR_V_M_1_USE_BUILTIN
#define CDN_MATH_OR_V_M_1 cdn_math_or_v_m_1_builtin

static ValueType *cdn_math_or_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_or_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] || x1;
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_M_1 */

#if defined(CDN_MATH_OR_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_OR_V_CWISE_M_1)
#define CDN_MATH_OR_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_OR_V_CWISE_M_1 cdn_math_or_v_cwise_m_1_builtin

static ValueType *cdn_math_or_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_or_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  ||  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_CWISE_M_1 */

#if defined(CDN_MATH_OR_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_OR_V_CWISE_1_M)
#define CDN_MATH_OR_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_OR_V_CWISE_1_M cdn_math_or_v_cwise_1_m_builtin

static ValueType *cdn_math_or_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_or_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  ||  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_CWISE_1_M */

#if defined(CDN_MATH_OR_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_OR_V_RWISE_M_1)
#define CDN_MATH_OR_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_OR_V_RWISE_M_1 cdn_math_or_v_rwise_m_1_builtin

static ValueType *cdn_math_or_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_or_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  ||  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_RWISE_M_1 */

#if defined(CDN_MATH_OR_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_OR_V_RWISE_1_M)
#define CDN_MATH_OR_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_OR_V_RWISE_1_M cdn_math_or_v_rwise_1_m_builtin

static ValueType *cdn_math_or_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_or_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  ||  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_OR_V_RWISE_1_M */

#if defined(CDN_MATH_AND_V_M_M_REQUIRED) && !defined(CDN_MATH_AND_V_M_M)
#define CDN_MATH_AND_V_M_M_USE_BUILTIN
#define CDN_MATH_AND_V_M_M cdn_math_and_v_m_m_builtin

static ValueType *cdn_math_and_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_and_v_m_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] && x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_M_M */

#if defined(CDN_MATH_AND_V_1_M_REQUIRED) && !defined(CDN_MATH_AND_V_1_M)
#define CDN_MATH_AND_V_1_M_USE_BUILTIN
#define CDN_MATH_AND_V_1_M cdn_math_and_v_1_m_builtin

static ValueType *cdn_math_and_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l);

static ValueType *cdn_math_and_v_1_m_builtin (ValueType *ret, ValueType x0, ValueType *x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0 && x1[i];
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_1_M */

#if defined(CDN_MATH_AND_V_M_1_REQUIRED) && !defined(CDN_MATH_AND_V_M_1)
#define CDN_MATH_AND_V_M_1_USE_BUILTIN
#define CDN_MATH_AND_V_M_1 cdn_math_and_v_m_1_builtin

static ValueType *cdn_math_and_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l);

static ValueType *cdn_math_and_v_m_1_builtin (ValueType *ret, ValueType *x0, ValueType x1, uint32_t l)
{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{
		ret[i] = x0[i] && x1;
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_M_1 */

#if defined(CDN_MATH_AND_V_CWISE_M_1_REQUIRED) && !defined(CDN_MATH_AND_V_CWISE_M_1)
#define CDN_MATH_AND_V_CWISE_M_1_USE_BUILTIN
#define CDN_MATH_AND_V_CWISE_M_1 cdn_math_and_v_cwise_m_1_builtin

static ValueType *cdn_math_and_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_and_v_cwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  &&  x1[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_CWISE_M_1 */

#if defined(CDN_MATH_AND_V_CWISE_1_M_REQUIRED) && !defined(CDN_MATH_AND_V_CWISE_1_M)
#define CDN_MATH_AND_V_CWISE_1_M_USE_BUILTIN
#define CDN_MATH_AND_V_CWISE_1_M cdn_math_and_v_cwise_1_m_builtin

static ValueType *cdn_math_and_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_and_v_cwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  &&  x0[r];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_CWISE_1_M */

#if defined(CDN_MATH_AND_V_RWISE_M_1_REQUIRED) && !defined(CDN_MATH_AND_V_RWISE_M_1)
#define CDN_MATH_AND_V_RWISE_M_1_USE_BUILTIN
#define CDN_MATH_AND_V_RWISE_M_1 cdn_math_and_v_rwise_m_1_builtin

static ValueType *cdn_math_and_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_and_v_rwise_m_1_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x0[i]  &&  x1[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_RWISE_M_1 */

#if defined(CDN_MATH_AND_V_RWISE_1_M_REQUIRED) && !defined(CDN_MATH_AND_V_RWISE_1_M)
#define CDN_MATH_AND_V_RWISE_1_M_USE_BUILTIN
#define CDN_MATH_AND_V_RWISE_1_M cdn_math_and_v_rwise_1_m_builtin

static ValueType *cdn_math_and_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns);

static ValueType *
cdn_math_and_v_rwise_1_m_builtin (ValueType *ret, ValueType *x0, ValueType *x1, uint32_t rows, uint32_t columns)
{
	uint32_t c;
	uint32_t i = 0;

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			ret[i] = x1[i]  &&  x0[c];
			++i;
		}
	}

	return ret;
}
#endif /* CDN_MATH_AND_V_RWISE_1_M */

#if defined(CDN_MATH_MAX_V_REQUIRED) && !defined(CDN_MATH_MAX_V)
#define CDN_MATH_MAX_V_USE_BUILTIN
#define CDN_MATH_MAX_V cdn_math_max_v_builtin

static ValueType cdn_math_max_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_max_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0];

	for (i = 1; i < l; ++i)
	{
		ret = CDN_MATH_MAX (ret, x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MAX_V */

#if defined(CDN_MATH_MIN_V_REQUIRED) && !defined(CDN_MATH_MIN_V)
#define CDN_MATH_MIN_V_USE_BUILTIN
#define CDN_MATH_MIN_V cdn_math_min_v_builtin

static ValueType cdn_math_min_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_min_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0];

	for (i = 1; i < l; ++i)
	{
		ret = CDN_MATH_MIN (ret, x0[i]);
	}

	return ret;
}
#endif /* CDN_MATH_MIN_V */

#if defined(CDN_MATH_HYPOT_V_REQUIRED) && !defined(CDN_MATH_HYPOT_V)
#define CDN_MATH_HYPOT_V_USE_BUILTIN
#define CDN_MATH_HYPOT_V cdn_math_hypot_v_builtin

static ValueType cdn_math_hypot_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_hypot_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0] * x0[0];

	for (i = 1; i < l; ++i)
	{
		ret += x0[i] * x0[i];
	}

	return CDN_MATH_SQRT (ret);
}
#endif /* CDN_MATH_HYPOT_V */

#if defined(CDN_MATH_SUM_V_REQUIRED) && !defined(CDN_MATH_SUM_V)
#define CDN_MATH_SUM_V_USE_BUILTIN
#define CDN_MATH_SUM_V cdn_math_sum_v_builtin

static ValueType cdn_math_sum_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_sum_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0];

	for (i = 1; i < l; ++i)
	{
		ret += x0[i];
	}

	return ret;
}
#endif /* CDN_MATH_SUM_V */

#if defined(CDN_MATH_PRODUCT_V_REQUIRED) && !defined(CDN_MATH_PRODUCT_V)
#define CDN_MATH_PRODUCT_V_USE_BUILTIN
#define CDN_MATH_PRODUCT_V cdn_math_product_v_builtin

static ValueType cdn_math_product_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_product_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0];

	for (i = 1; i < l; ++i)
	{
		ret *= x0[i];
	}

	return ret;
}
#endif /* CDN_MATH_PRODUCT_V */

#if defined(CDN_MATH_SQSUM_V_REQUIRED) && !defined(CDN_MATH_SQSUM_V)
#define CDN_MATH_SQSUM_V_USE_BUILTIN
#define CDN_MATH_SQSUM_V cdn_math_sqsum_v_builtin

static ValueType cdn_math_sqsum_v_builtin (ValueType *x0, uint32_t l);

static ValueType cdn_math_sqsum_v_builtin (ValueType *x0, uint32_t l)
{
	uint32_t i;
	ValueType ret = x0[0] * x0[0];

	for (i = 1; i < l; ++i)
	{
		ret += x0[i] * x0[i];
	}

	return ret;
}
#endif /* CDN_MATH_SQSUM_V */

#if defined(CDN_MATH_INDEX_V_REQUIRED) && !defined(CDN_MATH_INDEX_V)
#define CDN_MATH_INDEX_V_USE_BUILTIN
#define CDN_MATH_INDEX_V cdn_math_index_v_builtin


static ValueType *cdn_math_index_v_builtin (ValueType *ret,
                                            ValueType *x0,
                                            uint32_t  *indices,
                                            uint32_t   l);

static ValueType *
cdn_math_index_v_builtin (ValueType *ret,
                          ValueType *x0,
                          uint32_t  *indices,
                          uint32_t   l)
{{
	uint32_t i;

	for (i = 0; i < l; ++i)
	{{
		ret[i] = x0[indices[i]];
	}}

	return ret;
}}
#endif /* CDN_MATH_INDEX_V */

#if defined(CDN_MATH_VCAT_V_REQUIRED) && !defined(CDN_MATH_VCAT_V)
#define CDN_MATH_VCAT_V_USE_BUILTIN
#define CDN_MATH_VCAT_V cdn_math_vcat_v_builtin


static ValueType *cdn_math_vcat_v_builtin (ValueType *ret,
                                           ValueType *x0,
                                           ValueType *x1,
                                           uint32_t   rows1,
                                           uint32_t   rows2,
                                           uint32_t   columns);

static ValueType *
cdn_math_vcat_v_builtin (ValueType *ret,
                        ValueType *x0,
                        ValueType *x1,
                        uint32_t   rows1,
                        uint32_t   rows2,
                        uint32_t   columns)
{{
	uint32_t c;
	uint32_t i1 = 0;
	uint32_t i2 = 0;
	uint32_t ptr = 0;

	for (c = 0; c < columns; ++c)
	{{
		uint32_t r = 0;

		for (r = 0; r < rows1; ++r)
		{{
			ret[ptr++] = x0[i1++];
		}}

		for (r = 0; r < rows2; ++r)
		{{
			ret[ptr++] = x1[i2++];
		}}
	}}

	return ret;
}}
#endif /* CDN_MATH_VCAT_V */

#if defined(CDN_MATH_TRANSPOSE_REQUIRED) && !defined(CDN_MATH_TRANSPOSE)
#define CDN_MATH_TRANSPOSE_USE_BUILTIN
#define CDN_MATH_TRANSPOSE cdn_math_transpose_builtin


static ValueType cdn_math_transpose_builtin (ValueType x0);

static ValueType
cdn_math_tranpose_builtin (ValueType x0)
{{
	return x0;
}}
#endif /* CDN_MATH_TRANSPOSE */

#if defined(CDN_MATH_TRANSPOSE_V_REQUIRED) && !defined(CDN_MATH_TRANSPOSE_V)
#define CDN_MATH_TRANSPOSE_V_USE_BUILTIN
#define CDN_MATH_TRANSPOSE_V cdn_math_transpose_v_builtin


static ValueType *cdn_math_transpose_v_builtin (ValueType *ret,
                                                ValueType *x0,
                                                uint32_t   rows,
                                                uint32_t   columns);

static ValueType *
cdn_math_transpose_v_builtin (ValueType *ret,
                              ValueType *x0,
                              uint32_t   rows,
                              uint32_t   columns)
{{
	uint32_t c;
	uint32_t i = 0;
	uint32_t ptr = 0;

	for (c = 0; c < columns; ++c)
	{{
		uint32_t r;

		ptr = c;

		for (r = 0; r < rows; ++r)
		{{
			ret[ptr] = x0[i++];
			ptr += columns;
		}}
	}}

	return ret;
}}
#endif /* CDN_MATH_TRANSPOSE_V */

#if defined(CDN_MATH_MATRIX_MULTIPLY_V_REQUIRED) && !defined(CDN_MATH_MATRIX_MULTIPLY_V)
#define CDN_MATH_MATRIX_MULTIPLY_V_USE_BUILTIN
#define CDN_MATH_MATRIX_MULTIPLY_V cdn_math_matrix_multiply_v_builtin


static ValueType *cdn_math_matrix_multiply_v_builtin (ValueType *ret,
                                                      ValueType *x0,
                                                      ValueType *x1,
                                                      uint32_t   Rx0,
                                                      uint32_t   Cx0,
                                                      uint32_t   Cx1);

#ifndef ENABLE_BLAS

static ValueType *
cdn_math_matrix_multiply_v_no_blas_builtin (ValueType *ret,
                                            ValueType *x0,
                                            ValueType *x1,
                                            uint32_t   Rx0,
                                            uint32_t   Cx0,
                                            uint32_t   Cx1)
{{
	uint32_t c;
	uint32_t ptr = 0;

	for (c = 0; c < Cx1; ++c)
	{
		uint32_t r;

		for (r = 0; r < Rx0; ++r)
		{
			uint32_t i;
			uint32_t x0ptr = r;

			ret[ptr] = 0;

			for (i = 0; i < Cx0; ++i)
			{
				ret[ptr] += x0[x0ptr] * x1[i];
				x0ptr += Rx0;
			}

			++ptr;
		}

		x1 += Cx0;
	}

	return ret;
}}
#endif

#ifdef ENABLE_BLAS

#ifdef PLATFORM_OSX
#include <Accelerate/Accelerate.h>
#else
#include <cblas.h>
#endif

#define cblas_dgemmf cblas_sgemm

static ValueType *
cdn_math_matrix_multiply_v_blas_builtin (ValueType *ret,
                                         ValueType *x0,
                                         ValueType *x1,
                                         uint32_t   Rx0,
                                         uint32_t   Cx0,
                                         uint32_t   Cx1)
{{
	CDN_MATH_VALUE_TYPE_FUNC(cblas_dgemm)(CblasColMajor,
	                                      CblasNoTrans,
	                                      CblasNoTrans,
	                                      Rx0,
	                                      Cx1,
	                                      Cx0,
	                                      1,
	                                      x0,
	                                      Rx0,
	                                      x1,
	                                      Cx0,
	                                      0,
	                                      ret,
	                                      Rx0);

	return ret;
}}
#endif

static ValueType *
cdn_math_matrix_multiply_v_builtin (ValueType *ret,
                                    ValueType *x0,
                                    ValueType *x1,
                                    uint32_t   Rx0,
                                    uint32_t   Cx0,
                                    uint32_t   Cx1)
{{
#ifdef ENABLE_BLAS
	return cdn_math_matrix_multiply_v_blas_builtin (ret, x0, x1, Rx0, Cx0, Cx1);
#else
	return cdn_math_matrix_multiply_v_no_blas_builtin (ret, x0, x1, Rx0, Cx0, Cx1);
#endif
}}

#endif /* CDN_MATH_MATRIX_MULTIPLY_V */

#if defined(CDN_MATH_MATRIX_MULTIPLY_REQUIRED) && !defined(CDN_MATH_MATRIX_MULTIPLY)
#define CDN_MATH_MATRIX_MULTIPLY_USE_BUILTIN
#define CDN_MATH_MATRIX_MULTIPLY cdn_math_matrix_multiply_builtin


static ValueType cdn_math_matrix_multiply_builtin (ValueType *x0,
                                                   ValueType *x1,
                                                   uint32_t   Rx0,
                                                   uint32_t   Cx0,
                                                   uint32_t   Cx1);

static ValueType
cdn_math_matrix_multiply_builtin (ValueType *x0,
                                  ValueType *x1,
                                  uint32_t   Rx0,
                                  uint32_t   Cx0,
                                  uint32_t   Cx1)
{{
	ValueType retval;

	CDN_MATH_MATRIX_MULTIPLY_V (&retval, x0, x1, Rx0, Cx0, Cx1);
	return retval;
}}


#endif /* CDN_MATH_MATRIX_MULTIPLY */

#if defined(CDN_MATH_LINSOLVE_V_REQUIRED) && !defined(CDN_MATH_LINSOLVE_V)
#define CDN_MATH_LINSOLVE_V_USE_BUILTIN
#define CDN_MATH_LINSOLVE_V cdn_math_linsolve_v_builtin


static ValueType *cdn_math_linsolve_v_builtin (ValueType *ret,
                                               ValueType *A,
                                               ValueType *b,
                                               uint32_t   RA,
                                               uint32_t   CB,
                                               int64_t   *ipiv);

#ifndef ENABLE_LAPACK
static ValueType *
cdn_math_linsolve_v_no_lapack_builtin (ValueType *ret,
                                       ValueType *A,
                                       ValueType *b,
                                       uint32_t   RA,
                                       uint32_t   CB,
                                       int64_t   *ipiv)
{{
	#error("The linsolve function is not supported without LAPACK");
}}
#else
#ifdef PLATFORM_OSX
#include <vecLib/vecLib.h>
#else
#include <clapack.h>
#endif

#define fdgesv_ sgesv_

static ValueType *
cdn_math_linsolve_v_lapack_builtin (ValueType *ret,
                                    ValueType *A,
                                    ValueType *b,
                                    uint32_t   RA,
                                    uint32_t   CB,
                                    int64_t   *ipiv)
{{
	LP_ValueType *lpA = A;
	LP_ValueType *lpb = ret;
	LP_int lpRA = RA;
	LP_int lpCB = CB;
	LP_int info;
	LP_int *lpipiv = (LP_int *)ipiv;

	memcpy (ret, b, sizeof (ValueType) * RA * CB);

	CDN_MATH_VALUE_TYPE_FUNC(dgesv_) (&lpRA,
	                                  &lpCB,
	                                  lpA,
	                                  &lpRA,
	                                  lpipiv,
	                                  lpb,
	                                  &lpRA,
	                                  &info);

	return ret;
}}
#endif

static ValueType *
cdn_math_linsolve_v_builtin (ValueType *ret,
                             ValueType *A,
                             ValueType *b,
                             uint32_t   RA,
                             uint32_t   CB,
                             int64_t   *ipiv)
{{
#ifdef ENABLE_LAPACK
	return cdn_math_linsolve_v_lapack_builtin (ret, A, b, RA, CB, ipiv);
#else
	return cdn_math_linsolve_v_no_lapack_builtin (ret, A, b, RA, CB, ipiv);
#endif
}}

#endif /* CDN_MATH_LINSOLVE_V */

#if defined(CDN_MATH_QR_V_REQUIRED) && !defined(CDN_MATH_QR_V)
#define CDN_MATH_QR_V_USE_BUILTIN
#define CDN_MATH_QR_V cdn_math_qr_v_builtin


static ValueType *cdn_math_qr_v_builtin (ValueType *ret,
                                         ValueType *A,
                                         uint32_t   RA,
                                         uint32_t   CA,
                                         ValueType *tau,
                                         ValueType *work,
                                         uint32_t   lwork);

#ifndef ENABLE_LAPACK
static ValueType *
cdn_math_qr_v_no_lapack_builtin (ValueType *ret,
                                 ValueType *A,
                                 uint32_t   RA,
                                 uint32_t   CA,
                                 ValueType *tau,
                                 ValueType *work,
                                 uint32_t   lwork)
{{
	#error("The `qr' function is not supported without LAPACK");
}}
#else

#define fdgeqrf_ sgeqrf_
#define fdorgqr_ sorgqr_

static ValueType *
cdn_math_qr_v_lapack_builtin (ValueType *ret,
                              ValueType *A,
                              uint32_t   RA,
                              uint32_t   CA,
                              ValueType *tau,
                              ValueType *work,
                              uint32_t   lwork)
{{
	LP_ValueType *lpA     = A;
	LP_int        lpRA    = RA;
	LP_int        lpCA    = CA;
	LP_ValueType *lptau   = tau;
	LP_ValueType *lpwork  = work;
	LP_int        lplwork = lwork;
	LP_int        info;
	uint32_t      i;
	ValueType    *retptr;
	ValueType    *ptrA;

	CDN_MATH_VALUE_TYPE_FUNC(dgeqrf_) (&lpRA,
	                                   &lpCA,
	                                   lpA,
	                                   &lpRA,
	                                   lptau,
	                                   lpwork,
	                                   &lplwork,
	                                   &info);

	retptr = ret + RA * RA;
	ptrA = A;

	memset (retptr, 0, sizeof(ValueType) * RA * CA);

	// copy R to ret
	for (i = 1; i <= CA; ++i)
	{
		memcpy (retptr, ptrA, sizeof(ValueType) * i);

		retptr += RA;
		ptrA += RA;
	}

	// compute q
	dorgqr_ (&lpRA,
	         &lpRA,
	         &lpCA,
	         lpA,
	         &lpRA,
	         lptau,
	         lpwork,
	         &lplwork,
	         &info);

	// copy from A to ret
	memcpy (ret, A, sizeof(ValueType) * RA * RA);

	return ret;
}}
#endif

static ValueType *
cdn_math_qr_v_builtin (ValueType *ret,
                       ValueType *A,
                       uint32_t   RA,
                       uint32_t   CA,
                       ValueType *tau,
                       ValueType *work,
                       uint32_t   lwork)
{{
#ifdef ENABLE_LAPACK
	return cdn_math_qr_v_lapack_builtin (ret, A, RA, CA, tau, work, lwork);
#else
	return cdn_math_qr_v_no_lapack_builtin (ret, A, RA, CA, tau, work, lwork);
#endif
}}

#endif /* CDN_MATH_QR_V */

#if defined(CDN_MATH_SLINSOLVE_V_REQUIRED) && !defined(CDN_MATH_SLINSOLVE_V)
#define CDN_MATH_SLINSOLVE_V_USE_BUILTIN
#define CDN_MATH_SLINSOLVE_V cdn_math_slinsolve_v_builtin


static ValueType *cdn_math_slinsolve_v_builtin (ValueType *ret,
                                                ValueType *A,
                                                uint32_t   RA,
                                                ValueType *b,
                                                uint32_t   CB,
                                                ValueType *L);

static void
_cdn_math_slinsolve_factorize (ValueType *A,
                               ValueType *L,
                               int32_t    n)
{{
	int32_t k;

	for (k = n - 1; k >= 0; --k)
	{{
		int32_t i = (int32_t)L[k];
		int32_t ikk = k * (n + 1);

		while (i >= 0)
		{{
			int32_t j;
			ValueType a;
			int32_t iki;

			iki = k + i * n;

			// a = A_{iki} / A_{ikk}
			a = A[iki] / A[ikk];

			j = i;

			while (j >= 0)
			{{
				int32_t ijn = j * n;
				int32_t iij = i + ijn;
				int32_t ikj = k + ijn;

				// A_{ij} = A_{iij} - a A_{ikj}
				A[iij] -= a * A[ikj];

				j = (int32_t)L[j];
			}}

			// H_{ki} = a
			A[iki] = a;
			i = (int32_t)L[i];
		}}
	}}
}}

static void
_cdn_math_slinsolve_backsubs (ValueType *ptrA,
                              ValueType *ptrB,
                              ValueType *ptrL,
                              int32_t    n)
{{
	int32_t i;
	int32_t diag;

	diag = n * n - 1;

	// First solve for b = D^-1 L^-T b
	// see Sparce Factorization Algorithms, page 115
	for (i = n - 1; i >= 0; --i)
	{{
		int32_t j;

		j = (int32_t)ptrL[i];

		while (j >= 0)
		{{
			int32_t iij = i + j * n;

			// x_j = x_j - L_{ij} x_i
			ptrB[j] -= ptrA[iij] * ptrB[i];
			j = (int32_t)ptrL[j];
		}}

		// Apply D-1 from the diagonal elements if ptrA
		ptrB[i] /= ptrA[diag];
		diag -= n + 1;
	}}

	// Then finally solve for L^-1 b
	// see Sparce Factorization Algorithms, page 115
	for (i = 0; i < n; ++i)
	{{
		int32_t j;

		j = (int32_t)ptrL[i];

		while (j >= 0)
		{{
			int32_t iij = i + j * n;

			// x_i = x_i - L_{ij} x_j
			ptrB[i] -= ptrA[iij] * ptrB[j];
			j = (int32_t)ptrL[j];
		}}
	}}
}}

static ValueType *
cdn_math_slinsolve_v_builtin (ValueType *ret,
                              ValueType *A,
                              uint32_t   RA,
                              ValueType *b,
                              uint32_t   CB,
                              ValueType *L)
{{
	uint32_t i;
	ValueType *bptr;

	// Factorized A in place using LTDL factorization
	_cdn_math_slinsolve_factorize (A, L, RA);

	bptr = b;

	// Then backsubstitute
	for (i = 0; i < CB; ++i)
	{
		_cdn_math_slinsolve_backsubs (A, bptr, L, RA);
		bptr += RA;
	}

	memcpy (ret, b, sizeof (ValueType) * RA * CB);
	return ret;
}}

#endif /* CDN_MATH_SLINSOLVE_V */

#if defined(CDN_MATH_INVERSE_V_REQUIRED) && !defined(CDN_MATH_INVERSE_V)
#define CDN_MATH_INVERSE_V_USE_BUILTIN
#define CDN_MATH_INVERSE_V cdn_math_inverse_v_builtin


static ValueType *cdn_math_inverse_v_builtin (ValueType *ret,
                                              ValueType *A,
                                              uint32_t   RA,
                                              int64_t   *ipiv,
                                              ValueType *work,
                                              int32_t    lwork);

#ifndef ENABLE_LAPACK
static ValueType *
cdn_math_inverse_v_no_lapack_builtin (ValueType *ret,
                                      ValueType *A,
                                      uint32_t   RA,
                                      int64_t   *ipiv,
                                      ValueType *work,
                                      int32_t    lwork)
{{
	#error("The inv function is not supported without LAPACK");
}}
#else

#define fdgetrf_ sgetrf_
#define fdgetri_ sgetri_

static ValueType *
cdn_math_inverse_v_lapack_builtin (ValueType *ret,
                                   ValueType *A,
                                   uint32_t   RA,
                                   int64_t   *ipiv,
                                   ValueType *work,
                                   int32_t    lwork)
{{
	LP_ValueType *lpA = ret;
	LP_int lpRA = RA;
	LP_int *lpipiv = (LP_int *)ipiv;
	LP_ValueType *lpwork = work;
	LP_int lplwork = lwork;
	LP_int info;

	memcpy (ret, A, sizeof (ValueType) * RA * RA);

	CDN_MATH_VALUE_TYPE_FUNC(dgetrf_) (&lpRA,
	                                   &lpRA,
	                                   lpA,
	                                   &lpRA,
	                                   lpipiv,
	                                   &info);

	CDN_MATH_VALUE_TYPE_FUNC(dgetri_) (&lpRA,
	                                   lpA,
	                                   &lpRA,
	                                   lpipiv,
	                                   lpwork,
	                                   &lplwork,
	                                   &info);

	return ret;
}}
#endif

static ValueType *
cdn_math_inverse_v_builtin (ValueType *ret,
                             ValueType *A,
                             uint32_t   RA,
                             int64_t   *ipiv,
                             ValueType *work,
                             int32_t    lwork)
{{
#ifdef ENABLE_LAPACK
	return cdn_math_inverse_v_lapack_builtin (ret, A, RA, ipiv, work, lwork);
#else
	return cdn_math_inverse_v_no_lapack_builtin (ret, A, RA, ipiv, work, lwork);
#endif
}}

#endif /* CDN_MATH_INVERSE_V */

#if defined(CDN_MATH_PSEUDOINVERSE_V_REQUIRED) && !defined(CDN_MATH_PSEUDOINVERSE_V)
#define CDN_MATH_PSEUDOINVERSE_V_USE_BUILTIN
#define CDN_MATH_PSEUDOINVERSE_V cdn_math_pseudoinverse_v_builtin


static ValueType *cdn_math_pseudoinverse_v_builtin (ValueType *ret,
                                                     ValueType *A,
                                                     uint32_t   RA,
                                                     uint32_t   CA,
                                                     ValueType *B,
                                                     uint32_t   RB,
                                                     ValueType *S,
                                                     ValueType *work,
                                                     uint32_t   lwork,
                                                     int64_t   *iwork);

#ifndef ENABLE_LAPACK
static ValueType *
cdn_math_pseudoinverse_v_no_lapack_builtin (ValueType *ret,
                                             ValueType *A,
                                             uint32_t   RA,
                                             uint32_t   CA,
                                             ValueType *B,
                                             uint32_t   RB,
                                             ValueType *S,
                                             ValueType *work,
                                             uint32_t   lwork,
                                             int64_t   *iwork)
{{
	#error("The pinv function is not supported without LAPACK");
}}
#else

#define fdgelsd_ sgelsd_

static ValueType *
cdn_math_pseudoinverse_v_lapack_builtin (ValueType *ret,
                                          ValueType *A,
                                          uint32_t   RA,
                                          uint32_t   CA,
                                          ValueType *B,
                                          uint32_t   RB,
                                          ValueType *S,
                                          ValueType *work,
                                          uint32_t   lwork,
                                          int64_t   *iwork)
{{
	LP_ValueType *lpA     = A;
	LP_int        lpRA    = RA;
	LP_int        lpCA    = CA;
	LP_ValueType *lpB     = B;
	LP_int        lpRB    = RB;
	LP_ValueType *lpS     = S;
	LP_ValueType  rcond   = -1;
	LP_int        rank;
	LP_ValueType *lpwork  = work;
	LP_int        lplwork = lwork;
	LP_int       *lpiwork = (LP_int *)iwork;
	LP_int        info;
	uint32_t      i;
	ValueType    *retptr;

	CDN_MATH_VALUE_TYPE_FUNC(dgelsd_) (&lpRA,
	                                   &lpCA,
	                                   &lpRB,
	                                   lpA,
	                                   &lpRA,
	                                   lpB,
	                                   &lpRB,
	                                   lpS,
	                                   &rcond,
	                                   &rank,
	                                   lpwork,
	                                   &lplwork,
	                                   lpiwork,
	                                   &info);

	retptr = ret;

	// copy back to ret
	for (i = 0; i < RA; ++i)
	{{
		memcpy (retptr, B, sizeof (ValueType) * CA);

		retptr += CA;
		B += RB;
	}}

	return ret;
}}
#endif

static ValueType *
cdn_math_pseudoinverse_v_builtin (ValueType *ret,
                                  ValueType *A,
                                  uint32_t   RA,
                                  uint32_t   CA,
                                  ValueType *B,
                                  uint32_t   RB,
                                  ValueType *S,
                                  ValueType *work,
                                  uint32_t   lwork,
                                  int64_t   *iwork)
{{
#ifdef ENABLE_LAPACK
	return cdn_math_pseudoinverse_v_lapack_builtin (ret, A, RA, CA, B, RB, S, work, lwork, iwork);
#else
	return cdn_math_pseudoinverse_v_no_lapack_builtin (ret, A, RA, CA, B, RB, S, work, lwork, iwork);
#endif
}}

#endif /* CDN_MATH_PSEUDOINVERSE_V */

#if defined(CDN_MATH_DIAG_REQUIRED) && !defined(CDN_MATH_DIAG)
#define CDN_MATH_DIAG_USE_BUILTIN
#define CDN_MATH_DIAG cdn_math_diag_builtin


static void cdn_math_diag_builtin (ValueType x0);

static void
cdn_math_diag_builtin (ValueType x0)
{{
	return x0;
}}

#endif /* CDN_MATH_DIAG */

#if defined(CDN_MATH_DIAG_V_M_REQUIRED) && !defined(CDN_MATH_DIAG_V_M)
#define CDN_MATH_DIAG_V_M_USE_BUILTIN
#define CDN_MATH_DIAG_V_M cdn_math_diag_v_m_builtin


static ValueType *cdn_math_diag_v_m_builtin (ValueType *ret,
                                             ValueType *x0,
                                             uint32_t   n);

static ValueType *
cdn_math_diag_v_m_builtin (ValueType *ret,
                           ValueType *x0,
                           uint32_t   n)
{{
	uint32_t i;

	for (i = 0; i < n; ++i)
	{{
		ret[i] = *x0;
		x0 += n + 1;
	}}

	return ret;
}}
#endif /* CDN_MATH_DIAG_V_M */

#if defined(CDN_MATH_DIAG_V_V_REQUIRED) && !defined(CDN_MATH_DIAG_V_V)
#define CDN_MATH_DIAG_V_V_USE_BUILTIN
#define CDN_MATH_DIAG_V_V cdn_math_diag_v_v_builtin


static ValueType *cdn_math_diag_v_v_builtin (ValueType *ret,
                                             ValueType *x0,
                                             uint32_t   n);

static ValueType *
cdn_math_diag_v_v_builtin (ValueType *ret,
                           ValueType *x0,
                           uint32_t   n)
{{
	uint32_t i;
	ValueType *retptr = ret;

	for (i = 0; i < n; ++i)
	{{
		*retptr = x0[i];
		retptr += n + 1;
	}}

	return ret;
}}
#endif /* CDN_MATH_DIAG_V_V */

#if defined(CDN_MATH_TRIL_V_REQUIRED) && !defined(CDN_MATH_TRIL_V)
#define CDN_MATH_TRIL_V_USE_BUILTIN
#define CDN_MATH_TRIL_V cdn_math_tril_v_builtin


static ValueType *cdn_math_tril_v_builtin (ValueType *ret,
                                           ValueType *x0,
                                           uint32_t   rows,
                                           uint32_t   columns);

static ValueType *
cdn_math_tril_v_builtin (ValueType *ret,
                         ValueType *x0,
                         uint32_t   rows,
                         uint32_t   columns)
{{
	uint32_t c;
	ValueType *retptr = ret;

	for (c = 0; c < columns && c < rows; ++c)
	{{
		memcpy (retptr, x0, sizeof (ValueType) * (rows - c));

		retptr += rows + 1;
		x0 += rows + 1;
	}}

	return ret;
}}
#endif /* CDN_MATH_TRIL_V */

#if defined(CDN_MATH_TRIU_V_REQUIRED) && !defined(CDN_MATH_TRIU_V)
#define CDN_MATH_TRIU_V_USE_BUILTIN
#define CDN_MATH_TRIU_V cdn_math_triu_v_builtin


static ValueType *cdn_math_triu_v_builtin (ValueType *ret,
                                           ValueType *x0,
                                           uint32_t   rows,
                                           uint32_t   columns);

static ValueType *
cdn_math_triu_v_builtin (ValueType *ret,
                         ValueType *x0,
                         uint32_t   rows,
                         uint32_t   columns)
{{
	uint32_t c;
	ValueType *retptr = ret;

	for (c = 1; c <= columns; ++c)
	{{
		memcpy (retptr, x0, sizeof (ValueType) * (c > rows ? rows : c));

		retptr += rows;
		x0 += rows;
	}}

	return ret;
}}
#endif /* CDN_MATH_TRIU_V */

#if defined(CDN_MATH_CSUM_V_REQUIRED) && !defined(CDN_MATH_CSUM_V)
#define CDN_MATH_CSUM_V_USE_BUILTIN
#define CDN_MATH_CSUM_V cdn_math_csum_v_builtin


static ValueType *cdn_math_csum_v_builtin (ValueType *ret,
                                           ValueType *x0,
                                           uint32_t   rows,
                                           uint32_t   columns);

static ValueType *
cdn_math_csum_v_builtin (ValueType *ret,
                         ValueType *x0,
                         uint32_t   rows,
                         uint32_t   columns)
{
	uint32_t c;

	if (columns == 1)
	{
		memcpy (ret, x0, sizeof(ValueType) * rows);
		return ret;
	}

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		for (r = 0; r < rows; ++r)
		{
			if (c == 0)
			{
				ret[r] = *x0++;
			}
			else
			{
				ret[r] += *x0++;
			}
		}
	}

	return ret;
}
#endif /* CDN_MATH_CSUM_V */

#if defined(CDN_MATH_RSUM_V_REQUIRED) && !defined(CDN_MATH_RSUM_V)
#define CDN_MATH_RSUM_V_USE_BUILTIN
#define CDN_MATH_RSUM_V cdn_math_rsum_v_builtin


static ValueType *cdn_math_rsum_v_builtin (ValueType *ret,
                                           ValueType *x0,
                                           uint32_t   rows,
                                           uint32_t   columns);

static ValueType *
cdn_math_rsum_v_builtin (ValueType *ret,
                         ValueType *x0,
                         uint32_t   rows,
                         uint32_t   columns)
{
	uint32_t c;

	if (rows == 1)
	{
		memcpy (ret, x0, sizeof(ValueType) * columns);
		return ret;
	}

	for (c = 0; c < columns; ++c)
	{
		uint32_t r;

		ret[c] = 0;

		for (r = 0; r < rows; ++r)
		{
			ret[c] += *x0++;
		}
	}

	return ret;
}
#endif /* CDN_MATH_RSUM_V */
#endif /* CDN_RAWC_MATH_PROTOS */
